import asyncio
import concurrent.futures


def run_sync(coroutine):
    try:
        with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
            result = executor.submit(
                lambda coroutine_to_exec: asyncio.run(coroutine_to_exec), coroutine
            ).result()
        return result
    except RuntimeError as ex:
        print(repr(ex))
        return None



from supervisely.nn.benchmark.semantic_segmentation.base_vis_metric import (
    SemanticSegmVisMetric,
)
from supervisely.nn.benchmark.visualization.widgets import ChartWidget, MarkdownWidget


class KeyMetrics(SemanticSegmVisMetric):

    @property
    def md(self) -> MarkdownWidget:
        return MarkdownWidget(
            "markdown_header",
            "Key Metrics",
            text=self.vis_texts.markdown_key_metrics,
        )

    @property
    def chart(self) -> ChartWidget:
        return ChartWidget("base_metrics_chart", self.get_figure())

    def get_figure(self):
        import plotly.graph_objects as go  # pylint: disable=import-error

        fig = go.Figure()
        metrics = self.eval_result.mp.key_metrics().copy()
        metrics["mPixel accuracy"] = round(metrics["mPixel accuracy"] * 100, 2)
        fig.add_trace(
            go.Scatterpolar(
                r=list(metrics.values()) + [list(metrics.values())[0]],
                theta=list(metrics.keys()) + [list(metrics.keys())[0]],
                # fill="toself",
                hovertemplate="%{theta}: %{r:.2f}<extra></extra>",
            )
        )

        fig.update_layout(
            showlegend=False,
            polar=dict(
                radialaxis=dict(
                    range=[0, 100],
                    ticks="outside",
                ),
                angularaxis=dict(rotation=90, direction="clockwise"),
            ),
            dragmode=False,
            margin=dict(l=25, r=25, t=25, b=25),
            modebar=dict(
                remove=[
                    "zoom2d",
                    "pan2d",
                    "select2d",
                    "lasso2d",
                    "zoomIn2d",
                    "zoomOut2d",
                    "autoScale2d",
                    "resetScale2d",
                ]
            ),
        )
        return fig



from typing import Union

from supervisely.app.widgets import (
    Button,
    Card,
    Checkbox,
    Container,
    Editor,
    Field,
    Text,
)
from supervisely.nn.utils import RuntimeType


class HyperparametersSelector:
    title = "Hyperparameters"
    description = "Set hyperparameters for training"
    lock_message = "Select model to unlock"

    def __init__(self, hyperparameters: dict, app_options: dict = {}):
        self.display_widgets = []
        self.app_options = app_options

        # GUI Components
        self.editor = Editor(
            hyperparameters, height_lines=50, language_mode="yaml", auto_format=True
        )
        self.display_widgets.extend([self.editor])

        # Optional Model Benchmark
        if app_options.get("model_benchmark", True):
            # Model Benchmark
            self.run_model_benchmark_checkbox = Checkbox(
                content="Run Model Benchmark evaluation", checked=True
            )
            self.run_speedtest_checkbox = Checkbox(content="Run speed test", checked=True)

            self.model_benchmark_field = Field(
                title="Model Evaluation Benchmark",
                description=f"Generate evalutaion dashboard with visualizations and detailed analysis of the model performance after training. The best checkpoint will be used for evaluation. You can also run speed test to evaluate model inference speed.",
                content=Container([self.run_model_benchmark_checkbox, self.run_speedtest_checkbox]),
            )
            docs_link = '<a href="https://docs.supervisely.com/neural-networks/model-evaluation-benchmark/" target="_blank">documentation</a>'
            self.model_benchmark_learn_more = Text(
                f"Learn more about Model Benchmark in the {docs_link}.", status="info"
            )
            self.display_widgets.extend(
                [self.model_benchmark_field, self.model_benchmark_learn_more]
            )
        # -------------------------------- #

        # Optional Export Weights
        export_onnx_supported = app_options.get("export_onnx_supported", False)
        export_tensorrt_supported = app_options.get("export_tensorrt_supported", False)

        onnx_name = "ONNX"
        tensorrt_name = "TensorRT engine"
        export_runtimes = []
        export_runtime_names = []
        if export_onnx_supported:
            self.export_onnx_checkbox = Checkbox(content=f"Export to {onnx_name}")
            export_runtimes.append(self.export_onnx_checkbox)
            export_runtime_names.append(onnx_name)
        if export_tensorrt_supported:
            self.export_tensorrt_checkbox = Checkbox(content=f"Export to {tensorrt_name}")
            export_runtimes.append(self.export_tensorrt_checkbox)
            export_runtime_names.append(tensorrt_name)
        if export_onnx_supported or export_tensorrt_supported:
            export_field_description = ", ".join(export_runtime_names)
            runtime_container = Container(export_runtimes)
            self.export_field = Field(
                title="Export model",
                description=f"Export best checkpoint to the following formats: {export_field_description}.",
                content=runtime_container,
            )
            self.display_widgets.extend([self.export_field])
        # -------------------------------- #

        self.validator_text = Text("")
        self.validator_text.hide()
        self.button = Button("Select")
        self.display_widgets.extend([self.validator_text, self.button])
        # -------------------------------- #

        self.container = Container(self.display_widgets)
        self.card = Card(
            title=self.title,
            description=self.description,
            content=self.container,
            lock_message=self.lock_message,
            collapsable=app_options.get("collapsable", False),
        )
        self.card.lock()

    @property
    def widgets_to_disable(self) -> list:
        widgets = [self.editor]
        if self.app_options.get("model_benchmark", True):
            widgets.extend([self.run_model_benchmark_checkbox, self.run_speedtest_checkbox])
        if self.app_options.get("export_onnx_supported", False):
            widgets.append(self.export_onnx_checkbox)
        if self.app_options.get("export_tensorrt_supported", False):
            widgets.append(self.export_tensorrt_checkbox)
        return widgets

    def set_hyperparameters(self, hyperparameters: Union[str, dict]) -> None:
        self.editor.set_text(hyperparameters)

    def get_hyperparameters(self) -> dict:
        return self.editor.get_value()

    def get_model_benchmark_checkbox_value(self) -> bool:
        if self.app_options.get("model_benchmark", True):
            return self.run_model_benchmark_checkbox.is_checked()
        return False

    def set_model_benchmark_checkbox_value(self, is_checked: bool) -> bool:
        if is_checked:
            self.run_model_benchmark_checkbox.check()
        else:
            self.run_model_benchmark_checkbox.uncheck()

    def get_speedtest_checkbox_value(self) -> bool:
        if self.app_options.get("model_benchmark", True):
            return self.run_speedtest_checkbox.is_checked()
        return False

    def set_speedtest_checkbox_value(self, is_checked: bool) -> bool:
        if is_checked:
            self.run_speedtest_checkbox.check()
        else:
            self.run_speedtest_checkbox.uncheck()

    def toggle_mb_speedtest(self, is_checked: bool) -> None:
        if is_checked:
            self.run_speedtest_checkbox.show()
        else:
            self.run_speedtest_checkbox.hide()

    def get_export_onnx_checkbox_value(self) -> bool:
        if self.app_options.get("export_onnx_supported", False):
            return self.export_onnx_checkbox.is_checked()
        return False

    def set_export_onnx_checkbox_value(self, value: bool) -> None:
        if value:
            self.export_onnx_checkbox.check()
        else:
            self.export_onnx_checkbox.uncheck()

    def get_export_tensorrt_checkbox_value(self) -> bool:
        if self.app_options.get("export_tensorrt_supported", False):
            return self.export_tensorrt_checkbox.is_checked()
        return False

    def set_export_tensorrt_checkbox_value(self, value: bool) -> None:
        if value:
            self.export_tensorrt_checkbox.check()
        else:
            self.export_tensorrt_checkbox.uncheck()

    def is_export_required(self) -> bool:
        return self.get_export_onnx_checkbox_value() or self.get_export_tensorrt_checkbox_value()

    def validate_step(self) -> bool:
        return True



import random
from collections import defaultdict
from pathlib import Path

import supervisely.nn.benchmark.semantic_segmentation.text_templates as vis_texts
from supervisely.nn.benchmark.base_visualizer import BaseVisualizer
from supervisely.nn.benchmark.cv_tasks import CVTask
from supervisely.nn.benchmark.semantic_segmentation.vis_metrics.acknowledgement import (
    Acknowledgement,
)
from supervisely.nn.benchmark.semantic_segmentation.vis_metrics.classwise_error_analysis import (
    ClasswiseErrorAnalysis,
)
from supervisely.nn.benchmark.semantic_segmentation.vis_metrics.confusion_matrix import (
    ConfusionMatrix,
)
from supervisely.nn.benchmark.semantic_segmentation.vis_metrics.explore_predictions import (
    ExplorePredictions,
)
from supervisely.nn.benchmark.semantic_segmentation.vis_metrics.frequently_confused import (
    FrequentlyConfused,
)
from supervisely.nn.benchmark.semantic_segmentation.vis_metrics.iou_eou import (
    IntersectionErrorOverUnion,
)
from supervisely.nn.benchmark.semantic_segmentation.vis_metrics.key_metrics import (
    KeyMetrics,
)
from supervisely.nn.benchmark.semantic_segmentation.vis_metrics.model_predictions import (
    ModelPredictions,
)
from supervisely.nn.benchmark.semantic_segmentation.vis_metrics.overview import Overview
from supervisely.nn.benchmark.semantic_segmentation.vis_metrics.renormalized_error_ou import (
    RenormalizedErrorOverUnion,
)
from supervisely.nn.benchmark.semantic_segmentation.vis_metrics.speedtest import (
    Speedtest,
)
from supervisely.nn.benchmark.visualization.widgets import (
    ContainerWidget,
    MarkdownWidget,
    SidebarWidget,
)
from supervisely.project.project import Dataset, OpenMode, Project
from supervisely.project.project_meta import ProjectMeta


class SemanticSegmentationVisualizer(BaseVisualizer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.vis_texts = vis_texts
        self._widgets_created = False
        self.ann_opacity = 0.7

        diff_project_info, diff_dataset_infos, existed = self._get_or_create_diff_project()
        self.eval_result.diff_project_info = diff_project_info
        self.eval_result.diff_dataset_infos = diff_dataset_infos
        self.eval_result.matched_pair_data = {}

        self.gt_project_path = str(Path(self.workdir).parent / "gt_project")
        self.pred_project_path = str(Path(self.workdir).parent / "pred_project")

        self.eval_result.images_map = {}
        self.eval_result.images_by_class = defaultdict(set)
        if not existed:
            self._init_match_data()

        # set filtered project meta
        self.eval_result.filtered_project_meta = self._get_filtered_project_meta(self.eval_result)

        self._get_sample_data_for_gallery()

    @property
    def cv_task(self):
        return CVTask.SEMANTIC_SEGMENTATION

    def _create_widgets(self):
        # Modal Gellery
        self.diff_modal = self._create_diff_modal_table()
        self.explore_modal = self._create_explore_modal_table(
            click_gallery_id=self.diff_modal.id, hover_text="Compare with GT"
        )

        # Notifcation
        self.clickable_label = self._create_clickable_label()

        # overview
        overview = Overview(self.vis_texts, self.eval_result)
        self.header = overview.get_header(self.api.user.get_my_info().login)
        self.overview_md = overview.overview_md

        # key metrics
        key_metrics = KeyMetrics(self.vis_texts, self.eval_result)
        self.key_metrics_md = key_metrics.md
        self.key_metrics_chart = key_metrics.chart

        # explore predictions
        explore_predictions = ExplorePredictions(
            self.vis_texts, self.eval_result, self.explore_modal, self.diff_modal
        )
        self.explore_predictions_md = explore_predictions.md
        self.explore_predictions_gallery = explore_predictions.gallery(self.ann_opacity)

        # model predictions
        model_predictions = ModelPredictions(self.vis_texts, self.eval_result, self.diff_modal)
        self.model_predictions_md = model_predictions.md
        self.model_predictions_table = model_predictions.table

        # intersection over union
        iou_eou = IntersectionErrorOverUnion(self.vis_texts, self.eval_result)
        self.iou_eou_md = iou_eou.md
        self.iou_eou_chart = iou_eou.chart

        # renormalized error over union
        renorm_eou = RenormalizedErrorOverUnion(self.vis_texts, self.eval_result)
        self.renorm_eou_md = renorm_eou.md
        self.renorm_eou_chart = renorm_eou.chart

        # classwise error analysis
        classwise_error_analysis = ClasswiseErrorAnalysis(
            self.vis_texts, self.eval_result, self.explore_modal
        )
        self.classwise_error_analysis_md = classwise_error_analysis.md
        self.classwise_error_analysis_chart = classwise_error_analysis.chart

        # confusion matrix
        confusion_matrix = ConfusionMatrix(self.vis_texts, self.eval_result, self.explore_modal)
        self.confusion_matrix_md = confusion_matrix.md
        self.confusion_matrix_chart = confusion_matrix.chart

        # frequently confused
        frequently_confused = FrequentlyConfused(
            self.vis_texts, self.eval_result, self.explore_modal
        )
        self.frequently_confused_md = frequently_confused.md
        self.frequently_confused_chart = None
        if not frequently_confused.is_empty:
            self.frequently_confused_chart = frequently_confused.chart

        # Acknowledgement
        acknowledgement = Acknowledgement(self.vis_texts, self.eval_result)
        self.acknowledgement_md = acknowledgement.md

        # SpeedTest
        self.speedtest_present = False
        self.speedtest_multiple_batch_sizes = False
        speedtest = Speedtest(self.vis_texts, self.eval_result)
        if not speedtest.is_empty():
            self.speedtest_present = True
            self.speedtest_md_intro = speedtest.intro_md
            self.speedtest_intro_table = speedtest.intro_table
            if speedtest.multiple_batche_sizes():
                self.speedtest_multiple_batch_sizes = True
                self.speedtest_batch_inference_md = speedtest.batch_size_md
                self.speedtest_chart = speedtest.chart

        self._widgets_created = True

    def _create_layout(self):
        if not self._widgets_created:
            self._create_widgets()

        is_anchors_widgets = [
            # Overview
            (0, self.header),
            (1, self.overview_md),
            (1, self.key_metrics_md),
            (0, self.key_metrics_chart),
            (1, self.explore_predictions_md),
            (0, self.explore_predictions_gallery),
            (1, self.model_predictions_md),
            (0, self.model_predictions_table),
            (1, self.iou_eou_md),
            (0, self.iou_eou_chart),
            (1, self.renorm_eou_md),
            (0, self.renorm_eou_chart),
            (1, self.classwise_error_analysis_md),
            (0, self.clickable_label),
            (0, self.classwise_error_analysis_chart),
            (1, self.confusion_matrix_md),
            (0, self.clickable_label),
            (0, self.confusion_matrix_chart),
            (1, self.frequently_confused_md),
        ]
        if self.frequently_confused_chart is not None:
            is_anchors_widgets.append((0, self.clickable_label))
            is_anchors_widgets.append((0, self.frequently_confused_chart))
        if self.speedtest_present:
            is_anchors_widgets.append((1, self.speedtest_md_intro))
            is_anchors_widgets.append((0, self.speedtest_intro_table))
            if self.speedtest_multiple_batch_sizes:
                is_anchors_widgets.append((0, self.speedtest_batch_inference_md))
                is_anchors_widgets.append((0, self.speedtest_chart))

        is_anchors_widgets.append((0, self.acknowledgement_md))
        anchors = []
        for is_anchor, widget in is_anchors_widgets:
            if is_anchor:
                anchors.append(widget.id)

        sidebar = SidebarWidget(widgets=[i[1] for i in is_anchors_widgets], anchors=anchors)
        layout = ContainerWidget(
            widgets=[sidebar, self.explore_modal, self.diff_modal],
            name="main_container",
        )
        return layout

    def _init_match_data(self):
        gt_project = Project(self.gt_project_path, OpenMode.READ)
        pred_project = Project(self.pred_project_path, OpenMode.READ)
        diff_map = {ds.id: ds for ds in self.eval_result.diff_dataset_infos}
        pred_map = {ds.id: ds for ds in self.eval_result.pred_dataset_infos}

        def _get_full_name(ds_id: int, ds_id_map: dict):
            ds_info = ds_id_map[ds_id]
            if ds_info.parent_id is None:
                return ds_info.name
            return f"{_get_full_name(ds_info.parent_id, ds_id_map)}/{ds_info.name}"

        diff_dataset_name_map = {_get_full_name(i, diff_map): ds for i, ds in diff_map.items()}
        pred_dataset_name_map = {_get_full_name(i, pred_map): ds for i, ds in pred_map.items()}

        with self.pbar(
            message="Visualizations: Initializing match data", total=pred_project.total_items
        ) as p:
            for pred_dataset in pred_project.datasets:
                pred_dataset: Dataset
                gt_dataset: Dataset = gt_project.datasets.get(pred_dataset.name)
                try:
                    diff_dataset_info = diff_dataset_name_map[pred_dataset.name]
                    pred_dataset_info = pred_dataset_name_map[pred_dataset.name]
                except KeyError:
                    raise RuntimeError(
                        f"Difference project was not created properly. Dataset {pred_dataset.name} is missing"
                    )

                try:
                    for src_images in self.api.image.get_list_generator(
                        pred_dataset_info.id, force_metadata_for_links=False, batch_size=100
                    ):
                        dst_images = self.api.image.copy_batch_optimized(
                            pred_dataset_info.id,
                            src_images,
                            diff_dataset_info.id,
                            with_annotations=False,
                            skip_validation=True,
                        )
                        for diff_image_info in dst_images:
                            item_name = diff_image_info.name

                            gt_image_info = gt_dataset.get_image_info(item_name)
                            pred_image_info = pred_dataset.get_image_info(item_name)
                            gt_ann = gt_dataset.get_ann(item_name, gt_project.meta)
                            pred_ann = pred_dataset.get_ann(item_name, pred_project.meta)

                            self._update_match_data(
                                gt_image_info.id,
                                gt_image_info=gt_image_info,
                                pred_image_info=pred_image_info,
                                diff_image_info=diff_image_info,
                                gt_annotation=gt_ann,
                                pred_annotation=pred_ann,
                            )

                            assert item_name not in self.eval_result.images_map

                            self.eval_result.images_map[item_name] = gt_image_info.id

                            for label in pred_ann.labels:
                                self.eval_result.images_by_class[label.obj_class.name].add(
                                    gt_image_info.id
                                )

                        p.update(len(src_images))
                except Exception:
                    raise RuntimeError("Match data was not created properly")

    def _get_sample_data_for_gallery(self):
        # get sample images with annotations for visualization
        pred_ds = random.choice(self.eval_result.pred_dataset_infos)
        imgs = self.api.image.get_list(pred_ds.id, limit=9, force_metadata_for_links=False)
        anns = self.api.annotation.download_batch(
            pred_ds.id, [x.id for x in imgs], force_metadata_for_links=False
        )

        self.eval_result.sample_images = imgs
        self.eval_result.sample_anns = anns

    def _create_clickable_label(self):
        return MarkdownWidget(name="clickable_label", title="", text=self.vis_texts.clickable_label)

    def _get_filtered_project_meta(self, eval_result) -> ProjectMeta:
        remove_classes = []
        meta = eval_result.pred_project_meta.clone()
        meta = meta.merge(eval_result.gt_project_meta)
        if eval_result.classes_whitelist:
            for obj_class in meta.obj_classes:
                if obj_class.name not in eval_result.classes_whitelist:
                    if obj_class.name not in [eval_result.mp.bg_cls_name, "__bg__"]:
                        remove_classes.append(obj_class.name)
            if remove_classes:
                meta = meta.delete_obj_classes(remove_classes)
        return meta



from typing import Dict, Literal, Optional

from supervisely.app.content import DataJson, StateJson
from supervisely.app.widgets import Widget


class Dialog(Widget):
    """Dialog is a widget that allows to show a dialog window that contain any other widgets.
    It can be used to show a message to the user or to ask for confirmation.

    Read about it in `Developer Portal <https://developer.supervisely.com/app-development/widgets/layouts-and-containers/dialog>`_
        (including screenshots and examples).

    :param title: Dialog title
    :type title: str
    :param content: Dialog content
    :type content: Widget
    :param size: Dialog size, one of: tiny, small, large, full
    :type size: Literal["tiny", "small", "large", "full"]
    :param widget_id: An identifier of the widget.
    :type widget_id: str, optional

    :Usage example:
    .. code-block:: python

        from supervisely.app.widgets import Dialog, Input, Button

        dialog = Dialog(title="Dialog title", content=Input("Input"), size="large")
        dialog.show()
    """

    def __init__(
        self,
        title: Optional[str] = "",
        content: Optional[Widget] = None,
        size: Optional[Literal["tiny", "small", "large", "full"]] = "small",
        widget_id: Optional[str] = None,
    ):
        self._title = title
        self._content = content
        self._size = size
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict[str, str]:
        """Returns dictionary with widget data, which defines the appearance and behavior of the widget.

        Dictionary contains the following fields:
            - title: Dialog title
            - size: Dialog size, one of: tiny, small, large, full

        :return: Dictionary with widget data
        :rtype: Dict[str, str]
        """
        return {
            "title": self._title,
            "size": self._size,
        }

    def get_json_state(self) -> Dict[str, bool]:
        """Returns dictionary with widget state.

        Dictionary contains the following fields:
            - visible: Dialog visibility

        :return: Dictionary with widget state
        :rtype: Dict[str, bool]
        """
        return {
            "visible": False,
        }

    def show(self) -> None:
        """Shows the dialog."""
        StateJson()[self.widget_id]["visible"] = True
        StateJson().send_changes()

    def hide(self) -> None:
        """Hides the dialog."""
        StateJson()[self.widget_id]["visible"] = False
        StateJson().send_changes()

    @property
    def title(self) -> str:
        """Returns dialog title.

        :return: Dialog title
        :rtype: str
        """
        return self._title

    @title.setter
    def title(self, title: str) -> None:
        """Sets dialog title.

        :param title: Dialog title
        :type title: str
        """
        self._title = title
        DataJson()[self.widget_id]["title"] = title
        DataJson().send_changes()



# coding: utf-8
"""create/download/update :class:`Project<supervisely.project.project.Project>`"""

# docs
from __future__ import annotations

import os
from collections import defaultdict
from copy import deepcopy
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Literal,
    NamedTuple,
    Optional,
    Union,
)

from tqdm import tqdm

if TYPE_CHECKING:
    from pandas.core.frame import DataFrame

from datetime import datetime, timedelta

from supervisely import logger
from supervisely._utils import (
    abs_url,
    compare_dicts,
    compress_image_url,
    get_unix_timestamp,
    is_development,
)
from supervisely.annotation.annotation import TagCollection
from supervisely.annotation.obj_class import ObjClass
from supervisely.annotation.obj_class_collection import ObjClassCollection
from supervisely.annotation.tag_meta import TagMeta, TagValueType
from supervisely.api.module_api import (
    ApiField,
    CloneableModuleApi,
    RemoveableModuleApi,
    UpdateableModule,
)
from supervisely.io.json import dump_json_file, load_json_file
from supervisely.project.project_meta import ProjectMeta
from supervisely.project.project_meta import ProjectMetaJsonFields as MetaJsonF
from supervisely.project.project_settings import (
    ProjectSettings,
    ProjectSettingsJsonFields,
)
from supervisely.project.project_type import (
    _METADATA_SYSTEM_KEY,
    _METADATA_TIMESTAMP_KEY,
    _METADATA_VALIDATION_SCHEMA_KEY,
    _MULTISPECTRAL_TAG_NAME,
    _MULTIVIEW_TAG_NAME,
    ProjectType,
)


class ProjectNotFound(Exception):
    """ """

    pass


class ExpectedProjectTypeMismatch(Exception):
    """ """

    pass


class ProjectInfo(NamedTuple):
    """ """

    id: int
    name: str
    description: str
    size: int
    readme: str
    workspace_id: int
    images_count: int  # for compatibility with existing code
    items_count: int
    datasets_count: int
    created_at: str
    updated_at: str
    type: str
    reference_image_url: str
    custom_data: dict
    backup_archive: dict
    team_id: int
    settings: dict
    import_settings: dict
    version: dict

    @property
    def image_preview_url(self):
        if self.type in [str(ProjectType.POINT_CLOUDS), str(ProjectType.POINT_CLOUD_EPISODES)]:
            res = "https://user-images.githubusercontent.com/12828725/199022135-4161917c-05f8-4681-9dc1-b5e10ee8bb0f.png"
        else:
            res = self.reference_image_url
            if is_development():
                res = abs_url(res)
            res = compress_image_url(url=res, height=200)
        return res

    @property
    def url(self):
        res = f"projects/{self.id}/datasets"
        if is_development():
            res = abs_url(res)
        return res


class ProjectApi(CloneableModuleApi, UpdateableModule, RemoveableModuleApi):
    """
    API for working with :class:`Project<supervisely.project.project.Project>`. :class:`ProjectApi<ProjectApi>` object is immutable.

    :param api: API connection to the server
    :type api: Api
    :Usage example:

     .. code-block:: python

        import os
        from dotenv import load_dotenv

        import supervisely as sly

        # Load secrets and create API object from .env file (recommended)
        # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
        if sly.is_development():
            load_dotenv(os.path.expanduser("~/supervisely.env"))
        api = sly.Api.from_env()

        # Pass values into the API constructor (optional, not recommended)
        # api = sly.Api(server_address="https://app.supervise.ly", token="4r47N...xaTatb")

        project_id = 1951
        project_info = api.project.get_info_by_id(project_id)
    """

    @staticmethod
    def info_sequence():
        """
        NamedTuple ProjectInfo with API Fields containing information about Project.

        :Example:

         .. code-block:: python

            ProjectInfo(id=999,
                        name='Cat_breeds',
                        description='',
                        size='861069',
                        readme='',
                        workspace_id=58,
                        images_count=10,
                        items_count=10,
                        datasets_count=2,
                        created_at='2020-11-17T17:44:28.158Z',
                        updated_at='2021-03-01T10:51:57.545Z',
                        type='images',
                        reference_image_url='http://app.supervise.ly/h5un6l2bnaz1vj8a9qgms4-public/images/original/...jpg',
                        custom_data={},
                        backup_archive={},
                        team_id=2,
                        import_settings={}
                        )
        """
        return [
            ApiField.ID,
            ApiField.NAME,
            ApiField.DESCRIPTION,
            ApiField.SIZE,
            ApiField.README,
            ApiField.WORKSPACE_ID,
            ApiField.IMAGES_COUNT,  # for compatibility with existing code
            ApiField.ITEMS_COUNT,
            ApiField.DATASETS_COUNT,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
            ApiField.TYPE,
            ApiField.REFERENCE_IMAGE_URL,
            ApiField.CUSTOM_DATA,
            ApiField.BACKUP_ARCHIVE,
            ApiField.TEAM_ID,
            ApiField.SETTINGS,
            ApiField.IMPORT_SETTINGS,
            ApiField.VERSION,
        ]

    @staticmethod
    def info_tuple_name():
        """
        NamedTuple name - **ProjectInfo**.
        """
        return "ProjectInfo"

    def __init__(self, api):
        from supervisely.project.data_version import DataVersion

        CloneableModuleApi.__init__(self, api)
        UpdateableModule.__init__(self, api)
        self.version = DataVersion(api)

    def get_list(
        self,
        workspace_id: int,
        filters: Optional[List[Dict[str, str]]] = None,
        fields: List[str] = [],
    ) -> List[ProjectInfo]:
        """
        List of Projects in the given Workspace.

        *NOTE*: Version information is not available while getting list of projects.
        If you need version information, use :func:`get_info_by_id`.

        :param workspace_id: Workspace ID in which the Projects are located.
        :type workspace_id: int
        :param filters: List of params to sort output Projects.
        :type filters: List[dict], optional
        :param fields: The list of api fields which will be returned with the response.
        :type fields: List[str]

        :return: List of all projects with information for the given Workspace. See :class:`info_sequence<info_sequence>`
        :rtype: :class: `List[ProjectInfo]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            workspace_id = 58

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            project_list = api.project.get_list(workspace_id)
            print(project_list)
            # Output: [
            # ProjectInfo(id=861,
            #             name='Project_COCO',
            #             description='',
            #             size='22172241',
            #             readme='',
            #             workspace_id=58,
            #             images_count=6,
            #             items_count=6,
            #             datasets_count=1,
            #             created_at='2020-11-09T18:21:32.356Z',
            #             updated_at='2020-11-09T18:21:32.356Z',
            #             type='images',
            #             reference_image_url='http://78.46.75.100:38585/h5un6l2bnaz1vj8a9qgms4-public/images/original/...jpg',
            #             custom_data={},
            #             backup_archive={},
            #             import_settings={}
            #           ),
            # ProjectInfo(id=999,
            #             name='Cat_breeds',
            #             description='',
            #             size='861069',
            #             readme='',
            #             workspace_id=58,
            #             images_count=10,
            #             items_count=10,
            #             datasets_count=2,
            #             created_at='2020-11-17T17:44:28.158Z',
            #             updated_at='2021-03-01T10:51:57.545Z',
            #             type='images',
            #             reference_image_url='http://78.46.75.100:38585/h5un6l2bnaz1vj8a9qgms4-public/images/original/...jpg',
            #             custom_data={},
            #             backup_archive={},
            #             import_settings={}
            #           )
            # ]

            # Filtered Project list
            project_list = api.project.get_list(workspace_id, filters=[{ 'field': 'name', 'operator': '=', 'value': 'Cat_breeds'}])
            print(project_list)
            # Output: ProjectInfo(id=999,
            #                     name='Cat_breeds',
            #                     description='',
            #                     size='861069',
            #                     readme='',
            #                     workspace_id=58,
            #                     images_count=10,
            #                     items_count=10,
            #                     datasets_count=2,
            #                     created_at='2020-11-17T17:44:28.158Z',
            #                     updated_at='2021-03-01T10:51:57.545Z',
            #                     type='images',
            #                     reference_image_url='http://78.46.75.100:38585/h5un6l2bnaz1vj8a9qgms4-public/images/original/...jpg',
            #                     custom_data={},
            #                     backup_archive={},
            #                     import_settings={}
            #                   )
            # ]

        """
        if ApiField.VERSION in fields:
            fields.remove(ApiField.VERSION)
            logger.debug(
                "Project version information is not available while getting list of projects"
            )
        return self.get_list_all_pages(
            "projects.list",
            {
                ApiField.WORKSPACE_ID: workspace_id,
                ApiField.FILTER: filters or [],
                ApiField.FIELDS: fields,
            },
        )

    def get_info_by_id(
        self,
        id: int,
        expected_type: Optional[str] = None,
        raise_error: Optional[bool] = False,
    ) -> ProjectInfo:
        """
        Get Project information by ID.

        :param id: Project ID in Supervisely.
        :type id: int
        :param expected_type: Expected ProjectType.
        :type expected_type: ProjectType, optional
        :param raise_error: If True raise error if given name is missing in the Project, otherwise skips missing names.
        :type raise_error: bool, optional
        :raises: Error if type of project is not None and != expected type
        :return: Information about Project. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`ProjectInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            project_id = 1951

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            project_info = api.project.get_info_by_id(project_id)
            print(project_info)
            # Output: ProjectInfo(id=861,
            #                     name='fruits_annotated',
            #                     description='',
            #                     size='22172241',
            #                     readme='',
            #                     workspace_id=58,
            #                     images_count=6,
            #                     items_count=6,
            #                     datasets_count=1,
            #                     created_at='2020-11-09T18:21:32.356Z',
            #                     updated_at='2020-11-09T18:21:32.356Z',
            #                     type='images',
            #                     reference_image_url='http://78.46.75.100:38585/h5un6l2bnaz1vj8a9qgms4-public/images/original/...jpg',
            #                     custom_data={},
            #                     backup_archive={},
            #                     import_settings={}
            #                   )


        """
        info = self._get_info_by_id(id, "projects.info")
        self._check_project_info(info, id=id, expected_type=expected_type, raise_error=raise_error)
        return info

    def get_info_by_name(
        self,
        parent_id: int,
        name: str,
        expected_type: Optional[ProjectType] = None,
        raise_error: Optional[bool] = False,
    ) -> ProjectInfo:
        """
        Get Project information by name.

        Version information is not available while getting project by name.
        If you need version information, use :func:`get_info_by_id`.

        :param parent_id: Workspace ID.
        :type parent_id: int
        :param name: Project name.
        :type name: str
        :param expected_type: Expected ProjectType.
        :type expected_type: ProjectType, optional
        :param raise_error: If True raise error if given name is missing in the Project, otherwise skips missing names.
        :type raise_error: bool, optional
        :return: Information about Project. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`ProjectInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            project_info = api.project.get_info_by_name(58, "fruits_annotated")
            print(project_info)
            # Output: ProjectInfo(id=861,
            #                     name='fruits_annotated',
            #                     description='',
            #                     size='22172241',
            #                     readme='',
            #                     workspace_id=58,
            #                     images_count=6,
            #                     items_count=6,
            #                     datasets_count=1,
            #                     created_at='2020-11-09T18:21:32.356Z',
            #                     updated_at='2020-11-09T18:21:32.356Z',
            #                     type='images',
            #                     reference_image_url='http://78.46.75.100:38585/h5un6l2bnaz1vj8a9qgms4-public/images/original/...jpg',
            #                     custom_data={},
            #                     backup_archive={},
            #                     import_settings={}
            #                   )
        """

        fields = [
            x for x in self.info_sequence() if x not in (ApiField.ITEMS_COUNT, ApiField.SETTINGS)
        ]

        info = super().get_info_by_name(parent_id, name, fields)
        self._check_project_info(
            info, name=name, expected_type=expected_type, raise_error=raise_error
        )
        return info

    def _check_project_info(
        self,
        info,
        id: Optional[int] = None,
        name: Optional[str] = None,
        expected_type=None,
        raise_error=False,
    ):
        """
        Checks if a project exists with a given id and type of project == expected type
        :param info: project metadata information
        :param id: int
        :param name: str
        :param expected_type: type of data we expext to get info
        :param raise_error: bool
        """
        if raise_error is False:
            return

        str_id = ""
        if id is not None:
            str_id += "id: {!r} ".format(id)
        if name is not None:
            str_id += "name: {!r}".format(name)

        if info is None:
            raise ProjectNotFound("Project {} not found".format(str_id))
        if expected_type is not None and info.type != str(expected_type):
            raise ExpectedProjectTypeMismatch(
                "Project {!r} has type {!r}, but expected type is {!r}".format(
                    str_id, info.type, expected_type
                )
            )

    def get_meta(self, id: int, with_settings: bool = False) -> Dict:
        """
        Get ProjectMeta by Project ID.

        :param id: Project ID in Supervisely.
        :type id: int
        :param with_settings: Add settings field to the meta. By default False.
        :type with_settings: bool

        :return: ProjectMeta dict
        :rtype: :class:`dict`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            project_meta = api.project.get_meta(project_id)
            print(project_meta)
            # Output: {
            #     "classes":[
            #         {
            #             "id":22310,
            #             "title":"kiwi",
            #             "shape":"bitmap",
            #             "hotkey":"",
            #             "color":"#FF0000"
            #         },
            #         {
            #             "id":22309,
            #             "title":"lemon",
            #             "shape":"bitmap",
            #             "hotkey":"",
            #             "color":"#51C6AA"
            #         }
            #     ],
            #     "tags":[],
            #     "projectType":"images"
            # }
        """
        json_response = self._api.post("projects.meta", {"id": id}).json()

        if with_settings is True:
            json_settings = self.get_settings(id)
            mtag_name = None

            if json_settings.get("groupImagesByTagId") is not None:
                for tag in json_response["tags"]:
                    if tag["id"] == json_settings["groupImagesByTagId"]:
                        mtag_name = tag["name"]
                        break

            json_response[MetaJsonF.PROJECT_SETTINGS] = ProjectSettings(
                multiview_enabled=json_settings.get("groupImages", False),
                multiview_tag_name=mtag_name,
                multiview_tag_id=json_settings.get("groupImagesByTagId"),
                multiview_is_synced=json_settings.get("groupImagesSync", False),
                labeling_interface=json_settings.get(ProjectSettingsJsonFields.LABELING_INTERFACE),
            ).to_json()

        return json_response

    def clone_advanced(
        self,
        id,
        dst_workspace_id,
        dst_name,
        with_meta=True,
        with_datasets=True,
        with_items=True,
        with_annotations=True,
    ):
        """ """
        if not with_meta and with_annotations:
            raise ValueError(
                "with_meta parameter must be True if with_annotations parameter is True"
            )
        if not with_datasets and with_items:
            raise ValueError("with_datasets parameter must be True if with_items parameter is True")
        response = self._api.post(
            self._clone_api_method_name(),
            {
                ApiField.ID: id,
                ApiField.WORKSPACE_ID: dst_workspace_id,
                ApiField.NAME: dst_name,
                ApiField.INCLUDE: {
                    ApiField.CLASSES: with_meta,
                    ApiField.PROJECT_TAGS: with_meta,
                    ApiField.DATASETS: with_datasets,
                    ApiField.IMAGES: with_items,
                    ApiField.IMAGES_TAGS: with_items,
                    ApiField.ANNOTATION_OBJECTS: with_annotations,
                    ApiField.ANNOTATION_OBJECTS_TAGS: with_annotations,
                    ApiField.FIGURES: with_annotations,
                    ApiField.FIGURES_TAGS: with_annotations,
                },
            },
        )
        return response.json()[ApiField.TASK_ID]

    def create(
        self,
        workspace_id: int,
        name: str,
        type: ProjectType = ProjectType.IMAGES,
        description: Optional[str] = "",
        change_name_if_conflict: Optional[bool] = False,
    ) -> ProjectInfo:
        """
        Create Project with given name in the given Workspace ID.

        :param workspace_id: Workspace ID in Supervisely where Project will be created.
        :type workspace_id: int
        :param name: Project Name.
        :type name: str
        :param type: Type of created Project.
        :type type: ProjectType
        :param description: Project description.
        :type description: str
        :param change_name_if_conflict: Checks if given name already exists and adds suffix to the end of the name.
        :type change_name_if_conflict: bool, optional
        :return: Information about Project. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`ProjectInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            workspace_id = 8

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            new_proj = api.project.create(workspace_id, "fruits_test", sly.ProjectType.IMAGES)
            print(new_proj)
            # Output: ProjectInfo(id=1993,
            #                     name='fruits_test',
            #                     description='',
            #                     size='0',
            #                     readme='',
            #                     workspace_id=58,
            #                     images_count=None,
            #                     items_count=None,
            #                     datasets_count=None,
            #                     created_at='2021-03-11T09:28:42.585Z',
            #                     updated_at='2021-03-11T09:28:42.585Z',
            #                     type='images',
            #                     reference_image_url=None,
            #                     custom_data={},
            #                     backup_archive={},
            #                     import_settings={}
            #                   )

        """
        effective_name = self._get_effective_new_name(
            parent_id=workspace_id,
            name=name,
            change_name_if_conflict=change_name_if_conflict,
        )
        response = self._api.post(
            "projects.add",
            {
                ApiField.WORKSPACE_ID: workspace_id,
                ApiField.NAME: effective_name,
                ApiField.DESCRIPTION: description,
                ApiField.TYPE: str(type),
            },
        )
        return self._convert_json_info(response.json())

    def _get_update_method(self):
        """ """
        return "projects.editInfo"

    def update_meta(self, id: int, meta: Union[ProjectMeta, Dict]) -> ProjectMeta:
        """
        Updates given Project with given ProjectMeta.

        :param id: Project ID in Supervisely.
        :type id: int
        :param meta: ProjectMeta object or ProjectMeta in JSON format.
        :type meta: :class:`ProjectMeta` or dict

        :return: ProjectMeta
        :rtype: :class: `ProjectMeta`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            lemons_proj_id = 1951
            kiwis_proj_id = 1952

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            # Using ProjectMeta in JSON format

            project_meta_json = api.project.get_meta(lemons_proj_id)
            api.project.update_meta(kiwis_proj_id, project_meta_json)

            # Using ProjectMeta object

            project_meta_json = api.project.get_meta(lemons_proj_id)
            project_meta = sly.ProjectMeta.from_json(path_to_meta)
            api.project.update_meta(kiwis_proj_id, project_meta)

            # Using programmatically created ProjectMeta

            cat_class = sly.ObjClass("cat", sly.Rectangle, color=[0, 255, 0])
            scene_tag = sly.TagMeta("scene", sly.TagValueType.ANY_STRING)
            project_meta = sly.ProjectMeta(obj_classes=[cat_class], tag_metas=[scene_tag])
            api.project.update_meta(kiwis_proj_id, project_meta)

            # Update ProjectMeta from local `meta.json`
            from supervisely.io.json import load_json_file

            path_to_meta = "/path/project/meta.json"
            project_meta_json = load_json_file(path_to_meta)
            api.project.update_meta(kiwis_proj_id, project_meta)
        """

        m = meta
        if isinstance(meta, dict):
            m = ProjectMeta.from_json(meta)

        m.project_settings.validate(m)
        response = self._api.post(
            "projects.meta.update", {ApiField.ID: id, ApiField.META: m.to_json()}
        )
        try:
            tmp = ProjectMeta.from_json(data=response.json())
            m = tmp.clone(project_type=m.project_type, project_settings=m.project_settings)
        except KeyError:
            pass  # handle old instances <6.8.69: response.json()=={'success': True}

        if m.project_settings is not None:
            s = m.project_settings
            mtag_name = s.multiview_tag_name
            mtag_id = s.multiview_tag_id
            if mtag_name is None:
                mtag_name = m.get_tag_name_by_id(s.multiview_tag_id)

            if mtag_name is not None:  # (tag_id, tag_name)==(None, None) is OK but no group
                new_m = ProjectMeta.from_json(self.get_meta(id))
                group_tag = new_m.get_tag_meta(mtag_name)
                mtag_id = None if group_tag is None else group_tag.sly_id

            new_s = s.clone(
                multiview_tag_name=mtag_name,
                multiview_tag_id=mtag_id,
            )
            labeling_interface = new_s.labeling_interface
            m = m.clone(project_settings=new_s)
            settings_json = {
                "groupImages": new_s.multiview_enabled,
                "groupImagesByTagId": new_s.multiview_tag_id,
                "groupImagesSync": new_s.multiview_is_synced,
            }
            if labeling_interface is not None:
                settings_json[ProjectSettingsJsonFields.LABELING_INTERFACE] = labeling_interface
            self.update_settings(id, settings_json)

        return m

    def _clone_api_method_name(self):
        """ """
        return "projects.clone"

    def get_datasets_count(self, id: int) -> int:
        """
        Number of Datasets in the given Project by ID.

        :param id: Project ID in Supervisely.
        :type id: int
        :return: Number of Datasets in the given Project
        :rtype: :class:`int`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            project_id = 454

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            project_ds_count = api.project.get_datasets_count(project_id)
            print(project_ds_count)
            # Output: 4
        """
        datasets = self._api.dataset.get_list(id)
        return len(datasets)

    def get_images_count(self, id: int) -> int:
        """
        Number of images in the given Project by ID.

        :param id: Project ID in Supervisely.
        :type id: int
        :return: Number of images in the given Project
        :rtype: :class:`int`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            project_id = 454

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            project_imgs_count = api.project.get_images_count(project_id)
            print(project_imgs_count)
            # Output: 24
        """
        datasets = self._api.dataset.get_list(id, recursive=True)
        return sum([dataset.images_count for dataset in datasets])

    def _remove_api_method_name(self):
        """"""
        return "projects.remove"

    def merge_metas(self, src_project_id: int, dst_project_id: int) -> Dict:
        """
        Merges ProjectMeta from given Project to given destination Project.

        :param src_project_id: Source Project ID.
        :type src_project_id: int
        :param dst_project_id: Destination Project ID.
        :type dst_project_id: int
        :return: ProjectMeta dict
        :rtype: :class:`dict`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            lemons_proj_id = 1951
            kiwis_proj_id = 1980

            merged_projects = api.project.merge_metas(lemons_proj_id, kiwis_proj_id)
        """
        if src_project_id == dst_project_id:
            return self.get_meta(src_project_id)

        src_meta = ProjectMeta.from_json(self.get_meta(src_project_id))
        dst_meta = ProjectMeta.from_json(self.get_meta(dst_project_id))

        new_dst_meta = src_meta.merge(dst_meta)
        new_dst_meta_json = new_dst_meta.to_json()
        self.update_meta(dst_project_id, new_dst_meta.to_json())

        return new_dst_meta_json

    def get_activity(
        self, id: int, progress_cb: Optional[Union[tqdm, Callable]] = None
    ) -> DataFrame:
        """
        Get Project activity by ID.

        :param id: Project ID in Supervisely.
        :type id: int
        :return: `Pandas DataFrame <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html>`_
        :rtype: :class:`DataFrame`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            project_id = 1951

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            project_activity = api.project.get_activity(project_id)
            print(project_activity)
            # Output:    userId               action  ... tagId             meta
            #         0       7  annotation_duration  ...  None  {'duration': 1}
            #         1       7  annotation_duration  ...  None  {'duration': 2}
            #         2       7        create_figure  ...  None               {}
            #
            #         [3 rows x 18 columns]
        """
        import pandas as pd

        proj_info = self.get_info_by_id(id)
        workspace_info = self._api.workspace.get_info_by_id(proj_info.workspace_id)
        activity = self._api.team.get_activity(
            workspace_info.team_id, filter_project_id=id, progress_cb=progress_cb
        )
        df = pd.DataFrame(activity)
        return df

    def _convert_json_info(self, info: dict, skip_missing=True) -> ProjectInfo:
        """ """
        res = super()._convert_json_info(info, skip_missing=skip_missing)
        if res.reference_image_url is not None:
            res = res._replace(reference_image_url=res.reference_image_url)
        if res.items_count is None:
            res = res._replace(items_count=res.images_count)
        return ProjectInfo(**res._asdict())

    def get_stats(self, id: int) -> Dict:
        """
        Get Project stats by ID.

        :param id: Project ID in Supervisely.
        :type id: int
        :return: Project statistics
        :rtype: :class:`dict`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            project_id = 1951

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            project_stats = api.project.get_stats(project_id)
        """
        response = self._api.post("projects.stats", {ApiField.ID: id})
        return response.json()

    def url(self, id: int) -> str:
        """
        Get Project URL by ID.

        :param id: Project ID in Supervisely.
        :type id: int
        :return: Project URL
        :rtype: :class:`str`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            project_id = 1951

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            project_url = api.project.url(project_id)
            print(project_url)
            # Output: http://supervise.ly/projects/1951/datasets
        """
        res = f"projects/{id}/datasets"
        if is_development():
            res = abs_url(res)
        return res

    def update_custom_data(self, id: int, data: Dict, silent: bool = False) -> Dict:
        """
        Updates custom data of the Project by ID

        :param id: Project ID in Supervisely.
        :type id: int
        :param data: Custom data
        :type data: dict
        :param silent: determines whether the `updatedAt` timestamp should be updated or not, if False - update `updatedAt`
        :type silent: bool
        :return: Project information in dict format
        :rtype: :class:`dict`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            project_id = 1951
            custom_data = {1:2}

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            new_info = api.project.update_custom_data(project_id, custom_data)
        """
        if type(data) is not dict:
            raise TypeError("Meta must be dict, not {!r}".format(type(data)))
        response = self._api.post(
            "projects.editInfo",
            {ApiField.ID: id, ApiField.CUSTOM_DATA: data, ApiField.SILENT: silent},
        )
        return response.json()

    def get_custom_data(self, id: int) -> Dict[Any, Any]:
        """Returns custom data of the Project by ID.
        Custom data is a dictionary that can be used to store any additional information.

        :param id: Project ID in Supervisely.
        :type id: int
        :return: Custom data of the Project
        :rtype: :class:`dict`

        :Usage example:

        .. code-block:: python

            import supervisely as sly

            api = sly.Api.from_env()

            project_id = 123456

            custom_data = api.project.get_custom_data(project_id)

            print(custom_data) # Output: {'key': 'value'}
        """
        return self.get_info_by_id(id).custom_data

    def _get_system_custom_data(self, id: int) -> Dict[Any, Any]:
        """Returns system custom data of the Project by ID.
        System custom data is just a part of custom data that is used to store system information
        and obtained by the key `_METADATA_SYSTEM_KEY`.

        :param id: Project ID in Supervisely.
        :type id: int
        :return: System custom data of the Project
        :rtype: :class:`dict`

        :Usage example:

        .. code-block:: python

            import supervisely as sly

            api = sly.Api.from_env()

            project_id = 123456

            system_custom_data = api.project._get_system_custom_data(project_id)

            print(system_custom_data)
        """
        return self.get_info_by_id(id).custom_data.get(_METADATA_SYSTEM_KEY, {})

    def get_validation_schema(self, id: int, use_caching: bool = False) -> Optional[Dict[Any, Any]]:
        """Returns validation schema of the Project by ID.
        Validation schema is a dictionary that can be used to validate metadata of each entity in the project
        if corresnpoding schema is provided.
        If using caching, the schema will be loaded from the cache if available.
        Use cached version only in scenarios when the schema is not expected to change,
        otherwise it may lead to checks with outdated schema.

        :param id: Project ID in Supervisely.
        :type id: int
        :param use_caching: If True, uses cached version of the schema if available.
            NOTE: This may lead to checks with outdated schema. Use with caution.
            And only in scenarios when the schema is not expected to change.
        :return: Validation schema of the Project
        :rtype: :class:`dict`

        :Usage example:

        .. code-block:: python

            import supervisely as sly

            api = sly.Api.from_env()

            project_id = 123456

            validation_schema = api.project.get_validation_schema(project_id)

            print(validation_schema) # Output: {'key': 'Description of the field'}
        """
        SCHEMA_DIFF_THRESHOLD = 60 * 60  # 1 hour
        json_cache_filename = os.path.join(os.getcwd(), f"{id}_validation_schema.json")

        if use_caching:
            if os.path.isfile(json_cache_filename):
                try:
                    schema = load_json_file(json_cache_filename)
                    timestamp = schema.pop(_METADATA_TIMESTAMP_KEY, 0)

                    if get_unix_timestamp() - timestamp < SCHEMA_DIFF_THRESHOLD:
                        return schema
                except RuntimeError:
                    pass

        schema = self._get_system_custom_data(id).get(_METADATA_VALIDATION_SCHEMA_KEY)
        if schema and use_caching:
            schema_with_timestamp = deepcopy(schema)
            schema_with_timestamp[_METADATA_TIMESTAMP_KEY] = get_unix_timestamp()
            dump_json_file(schema_with_timestamp, json_cache_filename)

        return schema

    def _edit_validation_schema(
        self, id: int, schema: Optional[Dict[Any, Any]] = None
    ) -> Dict[Any, Any]:
        """Edits validation schema of the Project by ID.
        Do not use this method directly, use `set_validation_schema` or `remove_validation_schema` instead.

        :param id: Project ID in Supervisely.
        :type id: int
        :param schema: Validation schema to set. If None, removes validation schema.
        :type schema: dict, optional
        :return: Project information in dict format
        :rtype: :class:`dict`

        :Usage example:

        .. code-block:: python

            import supervisely as sly

            api = sly.Api.from_env()

            project_id = 123456

            schema = {'key': 'Description of the field'}

            api.project._edit_validation_schema(project_id, schema) #Set new validation schema.
            api.project._edit_validation_schema(project_id) #Remove validation schema.
        """
        custom_data = self.get_custom_data(id)
        system_data = custom_data.setdefault(_METADATA_SYSTEM_KEY, {})

        if not schema:
            system_data.pop(_METADATA_VALIDATION_SCHEMA_KEY, None)
        else:
            system_data[_METADATA_VALIDATION_SCHEMA_KEY] = schema
        return self.update_custom_data(id, custom_data)

    def set_validation_schema(self, id: int, schema: Dict[Any, Any]) -> Dict[Any, Any]:
        """Sets validation schema of the Project by ID.
        NOTE: This method will overwrite existing validation schema. To extend existing schema,
        use `get_validation_schema` first to get current schema, then update it and use this method to set new schema.

        :param id: Project ID in Supervisely.
        :type id: int
        :param schema: Validation schema to set.
        :type schema: dict
        :return: Project information in dict format
        :rtype: :class:`dict`

        :Usage example:

        .. code-block:: python

            import supervisely as sly

            api = sly.Api.from_env()

            project_id = 123456

            schema = {'key': 'Description of the field'}

            api.project.set_validation_schema(project_id, schema)
        """
        return self._edit_validation_schema(id, schema)

    def remove_validation_schema(self, id: int) -> Dict[Any, Any]:
        """Removes validation schema of the Project by ID.

        :param id: Project ID in Supervisely.
        :type id: int
        :return: Project information in dict format
        :rtype: :class:`dict`

        :Usage example:

        .. code-block:: python

            import supervisely as sly

            api = sly.Api.from_env()

            project_id = 123456

            api.project.remove_validation_schema(project_id)
        """
        return self._edit_validation_schema(id)

    def validate_entities_schema(
        self, id: int, strict: bool = False
    ) -> List[Dict[str, Union[id, str, List[str], List[Any]]]]:
        """Validates entities of the Project by ID using validation schema.
        Returns list of entities that do not match the schema.

        Example of the returned list:

        [
            {
                "entity_id": 123456,
                "entity_name": "image.jpg",
                "missing_fields": ["location"],
                "extra_fields": ["city.name"] <- Nested field (field "name" of the field "city")
            }
        ]

        :param id: Project ID in Supervisely.
        :type id: int
        :param strict: If strict is disabled, only checks if the entity has all the fields from the schema.
            Any extra fields in the entity will be ignored and will not be considered as an error.
            If strict is enabled, checks that the entity custom data is an exact match to the schema.
        :type strict: bool, optional
        :return: List of dictionaries with information about entities that do not match the schema.
        :rtype: :class:`List[Dict[str, Union[id, str, List[str], List[Any]]]`

        :Usage example:

        .. code-block:: python

            import supervisely as sly

            api = sly.Api.from_env()

            project_id = 123456

            incorrect_entities = api.project.validate_entities_schema(project_id)

            for entity in incorrect_entities:
                print(entity.id, entity.name) # Output: 123456, 'image.jpg'
        """
        validation_schema = self.get_validation_schema(id)
        if not validation_schema:
            raise ValueError("Validation schema is not set for this project.")

        info = self.get_info_by_id(id)
        listing_method = {
            # Mapping of project type to listing method.
            ProjectType.IMAGES.value: self._api.image.get_list,
        }
        custom_data_properties = {
            # Mapping of project type to custom data property name.
            # Can be different for different project types.
            ProjectType.IMAGES.value: "meta"
        }

        listing_method = listing_method.get(info.type)
        custom_data_property = custom_data_properties.get(info.type)

        if not listing_method or not custom_data_property:
            # TODO: Add support for other project types.
            raise NotImplementedError("Validation schema is not supported for this project type.")

        entities = listing_method(project_id=id)
        incorrect_entities = []

        for entity in entities:
            custom_data = getattr(entity, custom_data_property)
            missing_fields, extra_fields = compare_dicts(
                validation_schema, custom_data, strict=strict
            )
            if missing_fields or extra_fields:
                entry = {
                    "entity_id": entity.id,
                    "entity_name": entity.name,
                    "missing_fields": missing_fields,
                    "extra_fields": extra_fields,
                }
                incorrect_entities.append(entry)

        return incorrect_entities

    def get_settings(self, id: int) -> Dict[str, str]:
        info = self._get_info_by_id(id, "projects.info")
        return info.settings

    def update_settings(self, id: int, settings: Dict[str, str]) -> None:
        """
        Updates project wuth given project settings by id.

        :param id: Project ID
        :type id: int
        :param settings: Project settings
        :type settings: Dict[str, str]
        """
        self._api.post("projects.settings.update", {ApiField.ID: id, ApiField.SETTINGS: settings})

    def download_images_tags(
        self, id: int, progress_cb: Optional[Union[tqdm, Callable]] = None
    ) -> defaultdict:
        """
        Get matching tag names to ImageInfos.

        :param id: Project ID in Supervisely.
        :type id: int
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :return: Defaultdict matching tag names to ImageInfos
        :rtype: :class:`defaultdict`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            project_id = 8200
            tags_to_infos = api.project.download_images_tags(project_id)
            for tag_name in tags_to_infos:
                print(tag_name, tags_to_infos[tag_name])
            # Output:
            # train [ImageInfo(id=2389064, name='IMG_4451_JjH4WPkHlk.jpeg', link=None, hash='6EpjCL+lBdMBYo...
            # val [ImageInfo(id=2389066, name='IMG_1836.jpeg', link=None, hash='Si0WvJreU6pmrx1EDa1itkqqSkQkZFzNJSu...
        """
        # returns dict: tagname->images infos
        project_meta = self.get_meta(id)
        id_to_tagmeta = project_meta.tag_metas.get_id_mapping()
        tag2images = defaultdict(list)
        for dataset in self._api.dataset.get_list(id):
            ds_images = self._api.image.get_list(dataset.id)
            for img_info in ds_images:
                tags = TagCollection.from_api_response(
                    img_info.tags, project_meta.tag_metas, id_to_tagmeta
                )
                for tag in tags:
                    tag2images[tag.name].append(img_info)
                if progress_cb is not None:
                    progress_cb(1)
        return tag2images

    def images_grouping(self, id: int, enable: bool, tag_name: str, sync: bool = False) -> None:
        """Enables and disables images grouping by given tag name.

        :param id: Project ID, where images grouping will be enabled.
        :type id: int
        :param enable: if True groups images by given tag name, otherwise disables images grouping.
        :type enable: bool
        :param tag_name: Name of the tag. Images will be grouped by this tag.
        :type tag_name: str
        :param sync: Enable the syncronization views mode in the grouping settings. By default, `False`
        :type sync: bool
        """
        project_meta_json = self.get_meta(id)
        project_meta = ProjectMeta.from_json(project_meta_json)
        group_tag_meta = project_meta.get_tag_meta(tag_name)
        if group_tag_meta is None:
            raise RuntimeError(f"The group tag '{tag_name}' doesn't exist in the given project.")
        elif group_tag_meta.value_type != TagValueType.ANY_STRING:
            raise RuntimeError(
                f"The tag value type should be '{TagValueType.ANY_STRING}' for images grouping. The provided type: '{group_tag_meta.value_type}'"
            )
        group_tag_id = group_tag_meta.sly_id
        project_settings = {
            "groupImages": enable,
            "groupImagesByTagId": group_tag_id,
            "groupImagesSync": sync,
        }
        self.update_settings(id=id, settings=project_settings)

    def get_or_create(
        self,
        workspace_id: int,
        name: str,
        type: Optional[str] = ProjectType.IMAGES,
        description: Optional[str] = "",
    ) -> ProjectInfo:
        """Returns project info if project with given name exists in given workspace, otherwise creates new project
        and returns info about it.

        :param workspace_id: Workspace ID in which the Project will be searched or created.
        :type workspace_id: int
        :param name: name of the project to search or create
        :type name: str
        :param type: type of the project to create
        :type type: Optional[str], default ProjectType.IMAGES
        :param description: description of the project to create
        :type description: Optional[str]
        :return: ProjectInfo about found or created project
        :rtype: ProjectInfo
        :Usage example:

         .. code-block:: python

            import os
            from dotenv import load_dotenv

            import supervisely as sly

            # Load secrets and create API object from .env file (recommended)
            # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
            load_dotenv(os.path.expanduser("~/supervisely.env"))
            api = sly.Api.from_env()

            project_name = "my_project"
            workspace_id = 123
            project_info = api.project.get_or_create(workspace_id, project_name)
        """
        info = self.get_info_by_name(workspace_id, name)
        if info is None:
            info = self.create(workspace_id, name, type=type, description=description)
        return info

    def edit_info(
        self,
        id: int,
        name: Optional[str] = None,
        description: Optional[str] = None,
        readme: Optional[str] = None,
        custom_data: Optional[Dict[Any, Any]] = None,
        project_type: Optional[str] = None,
    ) -> ProjectInfo:
        """Edits the project info by given parameters.

        :param id: ID of the project to edit info
        :type id: int
        :param name: new name of the project
        :type name: Optional[str]
        :param description: new description of the project
        :type description: Optional[str]
        :param readme: new readme of the project
        :type readme: Optional[str]
        :param custom_data: new custom data of the project
        :type custom_data: Optional[Dict[Any, Any]]
        :param project_type: new type of the project
        :type project_type: Optional[str]
        :return: ProjectInfo of the edited project
        :rtype: ProjectInfo
        :raises ValueError: if no arguments are specified
        :raises ValueError: if invalid project type is specified
        :raises ValueError: if project with given id already has given type
        :raises ValueError: if conversion from current project type to given project type is not supported
        :Usage example:

         .. code-block:: python

            import os
            from dotenv import load_dotenv

            import supervisely as sly

            # Load secrets and create API object from .env file (recommended)
            # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
            load_dotenv(os.path.expanduser("~/supervisely.env"))
            api = sly.Api.from_env()

            project_id = 123
            new_name = "new_name"
            new_description = "new_description"
            project_info = api.project.edit_info(project_id, name=new_name, description=new_description)
        """
        if (
            name is None
            and description is None
            and readme is None
            and custom_data is None
            and project_type is None
        ):
            raise ValueError("one of the arguments has to be specified")

        body = {ApiField.ID: id}
        if name is not None:
            body[ApiField.NAME] = name
        if description is not None:
            body[ApiField.DESCRIPTION] = description
        if readme is not None:
            body[ApiField.README] = readme
        if custom_data is not None:
            body[ApiField.CUSTOM_DATA] = custom_data
        if project_type is not None:
            if isinstance(project_type, ProjectType):
                project_type = str(project_type)
            if project_type not in ProjectType.values():
                raise ValueError(f"project type must be one of: {ProjectType.values()}")
            project_info = self.get_info_by_id(id)
            current_project_type = project_info.type
            if project_type == current_project_type:
                raise ValueError(f"project with id {id} already has type {project_type}")
            if not (
                current_project_type == str(ProjectType.POINT_CLOUDS)
                and project_type == str(ProjectType.POINT_CLOUD_EPISODES)
            ):
                raise ValueError(
                    f"conversion from {current_project_type} to {project_type} is not supported "
                )
            body[ApiField.TYPE] = project_type

        response = self._api.post(self._get_update_method(), body)
        return self._convert_json_info(response.json())

    def pull_meta_ids(self, id: int, meta: ProjectMeta) -> None:
        """Updates given ProjectMeta with ids from server.

        :param id: Project ID
        :type id: int
        :param meta: ProjectMeta to update ids
        :type meta: ProjectMeta
        :Usage example:

         .. code-block:: python

            import os
            from dotenv import load_dotenv

            import supervisely as sly

            # Load secrets and create API object from .env file (recommended)
            # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
            load_dotenv(os.path.expanduser("~/supervisely.env"))
            api = sly.Api.from_env()

            project_id = 123
            # We already have ProjectMeta and now we want to update ids in it
            # from server
            meta: sly.ProjectMeta

            api.project.pull_meta_ids(project_id, meta)
        """
        # to update ids in existing project meta
        meta_json = self.get_meta(id)
        server_meta = ProjectMeta.from_json(meta_json)
        meta.obj_classes.refresh_ids_from(server_meta.obj_classes)
        meta.tag_metas.refresh_ids_from(server_meta.tag_metas)

    def move(self, id: int, workspace_id: int) -> None:
        """
        Move project between workspaces within current team.

        :param id: Project ID
        :type id: int
        :param workspace_id: Workspace ID the project will move in
        :type workspace_id: int
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            workspace_id = 688
            project_id = 17173

            api.project.move(id=project_id, workspace_id=workspace_id)
        """
        self._api.post(
            "projects.workspace.set", {ApiField.ID: id, ApiField.WORKSPACE_ID: workspace_id}
        )

    def archive_batch(
        self, ids: List[int], archive_urls: List[str], ann_archive_urls: Optional[List[str]] = None
    ) -> None:
        """
        Archive Projects by ID and save backup URLs in Project info for every Project.

        :param ids: Project IDs in Supervisely.
        :type ids: List[int]
        :param archive_urls: Shared URLs of files backup on Dropbox.
        :type archive_urls: List[str]
        :param ann_archive_urls: Shared URLs of annotations backup on Dropbox.
        :type ann_archive_urls: List[str], optional
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

        .. code-block:: python

            import supervisely as sly

            ids = [18464, 18461]
            archive_urls = ['https://www.dropbox.com/...', 'https://www.dropbox.com/...']

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            api.project.archive_batch(ids, archive_urls, ann_archive_urls)
        """
        if len(ids) != len(archive_urls):
            raise ValueError(
                "The list with Project IDs must have the same length as the list with URLs for archives"
            )
        for id, archive_url, ann_archive_url in zip(
            ids, archive_urls, ann_archive_urls or [None] * len(ids)
        ):
            request_params = {
                ApiField.ID: id,
                ApiField.ARCHIVE_URL: archive_url,
            }
            if ann_archive_url is not None:
                request_params[ApiField.ANN_ARCHIVE_URL] = ann_archive_url

            self._api.post("projects.remove.permanently", {ApiField.PROJECTS: [request_params]})

    def archive(self, id: int, archive_url: str, ann_archive_url: Optional[str] = None) -> None:
        """
        Archive Project by ID and save backup URLs in Project info.

        :param id: Project ID in Supervisely.
        :type id: int
        :param archive_url: Shared URL of files backup on Dropbox.
        :type archive_url: str
        :param ann_archive_url: Shared URL of annotations backup on Dropbox.
        :type ann_archive_url: str, optional
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

        .. code-block:: python

            import supervisely as sly

            id = 18464
            archive_url = 'https://www.dropbox.com/...'

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            api.project.archive(id, archive_url, ann_archive_url)
        """
        if ann_archive_url is None:
            self.archive_batch([id], [archive_url])
        else:
            self.archive_batch([id], [archive_url], [ann_archive_url])

    def get_archivation_list(
        self,
        to_day: Optional[int] = None,
        from_day: Optional[int] = None,
        skip_exported: Optional[bool] = None,
        sort: Optional[
            Literal[
                "id",
                "title",
                "size",
                "createdAt",
                "updatedAt",
            ]
        ] = None,
        sort_order: Optional[Literal["asc", "desc"]] = None,
        account_type: Optional[str] = None,
    ) -> List[ProjectInfo]:
        """
        List of all projects in all available workspaces that can be archived.

        :param to_day: Sets the number of days from today. If the project has not been updated during this period, it will be added to the list.
        :type to_day: int, optional
        :param from_day: Sets the number of days from today. If the project has not been updated before this period, it will be added to the list.
        :type from_day: int, optional
        :param skip_exported: Determines whether to skip already archived projects.
        :type skip_exported: bool, optional.
        :param sort: Specifies by which parameter to sort the project list.
        :type sort: Optional[Literal["id", "title", "size", "createdAt", "updatedAt"]]
        :param sort_order: Determines which value to list from.
        :type sort_order: Optional[Literal["asc", "desc"]]
        :return: List of all projects with information. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[ProjectInfo]`
        :Usage example:

        .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            project_list = api.project.get_archivation_list()
            print(project_list)
            # Output: [
            # ProjectInfo(id=861,
            #             name='Project_COCO'
            #             size='22172241',
            #             workspace_id=58,
            #             created_at='2020-11-09T18:21:32.356Z',
            #             updated_at='2020-11-09T18:21:32.356Z',
            #             type='images',
            #             ...
            #             ),
            # ProjectInfo(id=777,
            #             name='Trucks',
            #             size='76154769',
            #             workspace_id=58,
            #             created_at='2021-07-077T17:44:28.158Z',
            #             updated_at='2023-07-15T12:33:45.747Z',
            #             type='images',)
            # ]

            # Project list for desired date range
            project_list = api.project.get_archivation_list(to_day=2)
            print(project_list)
            # Output: ProjectInfo(id=777,
            #                     name='Trucks',
            #                     size='76154769',
            #                     workspace_id=58,
            #                     created_at='2021-07-077T17:44:28.158Z',
            #                     updated_at='2023-07-15T12:33:45.747Z',
            #                     type='images',
            #                     ...
            #                     )
            # ]

        """
        kwargs = {}

        filters = []
        if from_day is not None:
            date = (datetime.utcnow() - timedelta(days=from_day)).strftime("%Y-%m-%dT%H:%M:%SZ")
            filer_from = {
                ApiField.FIELD: ApiField.UPDATED_AT,
                "operator": ">=",
                ApiField.VALUE: date,
            }
            filters.append(filer_from)
        if to_day is not None:
            date = (datetime.utcnow() - timedelta(days=to_day)).strftime("%Y-%m-%dT%H:%M:%SZ")
            filer_to = {
                ApiField.FIELD: ApiField.UPDATED_AT,
                "operator": "<=",
                ApiField.VALUE: date,
            }
            filters.append(filer_to)
        if len(filters) != 0:
            kwargs["filters"] = filters

        if skip_exported is None:
            kwargs["skip_exported"] = skip_exported

        if sort is not None:
            kwargs["sort"] = sort

        if sort_order is not None:
            kwargs["sort_order"] = sort_order

        if account_type is not None:
            kwargs["account_type"] = account_type

        response = self.get_list_all(**kwargs)

        return response.get("entities")

    def check_imageset_backup(self, id: int) -> Optional[Dict]:
        """
        Check if a backup of the project image set exists. If yes, it returns a link to the archive.

        :param id: Project ID
        :type id: int
        :return: dict with shared URL of files backup or None
        :rtype: Dict, optional
        :Usage example:

         .. code-block:: python

            import os
            from dotenv import load_dotenv

            import supervisely as sly

            # Load secrets and create API object from .env file (recommended)
            # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
            if sly.is_development():
               load_dotenv(os.path.expanduser("~/supervisely.env"))
            api = sly.Api.from_env()

            # Pass values into the API constructor (optional, not recommended)
            # api = sly.Api(server_address="https://app.supervise.ly", token="4r47N...xaTatb")

            response = check_imageset_backup(project_id)
            archive_url = response['imagesArchiveUrl']

        """
        response = self._api.get("projects.images.get-backup-archive", {ApiField.ID: id})

        return response.json()

    def append_classes(self, id: int, classes: Union[List[ObjClass], ObjClassCollection]) -> None:
        """
        Append new classes to given Project.

        :param id: Project ID in Supervisely.
        :type id: int
        :param classes: New classes
        :type classes: :class: ObjClassCollection or List[ObjClass]
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            proj_id = 28145
            lung_obj_class = sly.ObjClass("lung", sly.Mask3D)
            api.project.append_classes(proj_id, [lung_obj_class])
        """
        meta_json = self.get_meta(id)
        meta = ProjectMeta.from_json(meta_json)
        meta = meta.add_obj_classes(classes)
        self.update_meta(id, meta)

    def _set_custom_grouping_settings(self, id: int, group_images: bool, tag_name: str, sync: bool):
        """Sets the project settings for custom grouping.

        :param id: Project ID to set custom grouping settings.
        :type id: int
        :param group_images: if True enables images grouping by tag
        :type group_images: bool
        :param tag_name: Name of the tag. Images will be grouped by this tag
        :type tag_name: str
        :param sync: if True images will have synchronized view and labeling
        :type sync: bool
        """
        meta = ProjectMeta.from_json(self.get_meta(id, with_settings=True))
        existing_tag_meta = meta.get_tag_meta(tag_name)
        if existing_tag_meta is not None:
            if existing_tag_meta.value_type != TagValueType.ANY_STRING:
                raise ValueError(
                    f"Tag '{tag_name}' should have value type 'any_string', "
                    f"but got '{existing_tag_meta.value_type}' value type."
                )
        else:
            new_tag_meta = TagMeta(tag_name, TagValueType.ANY_STRING)
            project_meta = meta.add_tag_meta(new_tag_meta)
            self.update_meta(id, project_meta)
        self.images_grouping(id, enable=group_images, tag_name=tag_name, sync=sync)

    def set_multispectral_settings(self, project_id: int) -> None:
        """Sets the project settings for multispectral images.
        Images will be grouped by tag and have synchronized view and labeling.

        :param project_id: Project ID to set multispectral settings.
        :type project_id: int
        :Usage example:

         .. code-block:: python

            import os
            from dotenv import load_dotenv

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'

            # Load secrets and create API object from .env file (recommended)
            # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
            load_dotenv(os.path.expanduser("~/supervisely.env"))
            api = sly.Api.from_env()

            api.project.set_multispectral_settings(project_id=123)
        """

        self._set_custom_grouping_settings(
            id=project_id,
            group_images=True,
            tag_name=_MULTISPECTRAL_TAG_NAME,
            sync=True,
        )

    def set_multiview_settings(self, project_id: int) -> None:
        """Sets the project settings for multiview images.
        Images will be grouped by tag and have synchronized view and labeling.

        :param project_id: Project ID to set multiview settings.
        :type project_id: int
        :Usage example:

         .. code-block:: python

            import os
            from dotenv import load_dotenv

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'

            # Load secrets and create API object from .env file (recommended)
            # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
            load_dotenv(os.path.expanduser("~/supervisely.env"))
            api = sly.Api.from_env()

            api.project.set_multiview_settings(project_id=123)
        """

        self._set_custom_grouping_settings(
            id=project_id,
            group_images=True,
            tag_name=_MULTIVIEW_TAG_NAME,
            sync=False,
        )

    def remove_permanently(
        self, ids: Union[int, List], batch_size: int = 50, progress_cb=None
    ) -> List[dict]:
        """
        !!! WARNING !!!
        Be careful, this method deletes data from the database, recovery is not possible.

        Delete permanently projects with given IDs from the Supervisely server.
        All project IDs must belong to the same team.
        Therefore, it is necessary to sort IDs before calling this method.

        :param ids: IDs of projects in Supervisely.
        :type ids: Union[int, List]
        :param batch_size: The number of entities that will be deleted by a single API call. This value must be in the range 1-50 inclusive, if you set a value out of range it will automatically adjust to the boundary values.
        :type batch_size: int, optional
        :param progress_cb: Function for control delete progress.
        :type progress_cb: Callable, optional
        :return: A list of response content in JSON format for each API call.
        :rtype: List[dict]
        """
        if batch_size > 50:
            batch_size = 50
        elif batch_size < 1:
            batch_size = 1

        if isinstance(ids, int):
            projects = [{ApiField.ID: ids}]
        else:
            projects = [{ApiField.ID: id} for id in ids]

        batches = [projects[i : i + batch_size] for i in range(0, len(projects), batch_size)]
        responses = []
        for batch in batches:
            request_body = {
                ApiField.PROJECTS: batch,
                ApiField.PRESERVE_PROJECT_CARD: False,
            }
            response = self._api.post("projects.remove.permanently", request_body)
            if progress_cb is not None:
                progress_cb(len(batch))
            responses.append(response.json())
        return responses

    def get_list_all(
        self,
        filters: Optional[List[Dict[str, str]]] = None,
        skip_exported: Optional[bool] = True,
        sort: Optional[str] = None,
        sort_order: Optional[str] = None,
        per_page: Optional[int] = None,
        page: Union[int, Literal["all"]] = "all",
        account_type: Optional[str] = None,
    ) -> dict:
        """
        List all available projects from all available teams for the user that match the specified filtering criteria.

        :param filters: List of parameters for filtering the available Projects.
                        Every Dict must consist of keys:
                        - 'field': Takes values 'id', 'projectId', 'workspaceId', 'groupId', 'createdAt', 'updatedAt', 'type'
                        - 'operator': Takes values '=', 'eq', '!=', 'not', 'in', '!in', '>', 'gt', '>=', 'gte', '<', 'lt', '<=', 'lte'
                        - 'value': Takes on values according to the meaning of 'field' or null
        :type filters: List[Dict[str, str]], optional

        :param skip_exported: Determines whether to skip archived projects.
        :type skip_exported: bool, optional.

        :param sort: Specifies by which parameter to sort the project list.
                        Takes values 'id', 'name', 'size', 'createdAt', 'updatedAt'
        :type sort: str, optional

        :param sort_order: Determines which value to list from.
        :type sort_order: str, optional

        :param per_page: Number of first items found to be returned.
                        'None' will return the first page with a default size of 20000 projects.
        :type per_page: int, optional

        :param page: Page number, used to retrieve the following items if the number of them found is more than per_page.
                     The default value is 'all', which retrieves all available projects.
                     'None' will return the first page with projects, the amount of which is set in param 'per_page'.
        :type page: Union[int, Literal["all"]], optional

        :param account_type: (Deprecated) Type of user account
        :type account_type: str, optional

        :return: Search response information and 'ProjectInfo' of all projects that are searched by a given criterion.
        :rtype: dict

        :Usage example:

        .. code-block:: python

            import supervisely as sly
            import os

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            filter_1 = {
                "field": "updatedAt",
                "operator": "<",
                "value": "2023-12-03T14:53:00.952Z"
            }
            filter_2 = {
                "field": "updatedAt",
                "operator": ">",
                "value": "2023-04-03T14:53:00.952Z"
            }
            filters = [filter_1, filter_2]
            projects = api.projects.get_list_all(filters, True)
            print(projects)
            # Output:
            # {
            #     "total": 2,
            #     "perPage": 20000,
            #     "pagesCount": 1,
            #     "entities": [ ProjectInfo(id = 22,
            #                       name = 'lemons_annotated',
            #                       description = None,
            #                       size = '861069',
            #                       readme = None,
            #                       workspace_id = 2,
            #                       images_count = None,
            #                       items_count = None,
            #                       datasets_count = None,
            #                       created_at = '2020-04-03T13:43:24.000Z',
            #                       updated_at = '2020-04-03T14:53:00.952Z',
            #                       type = 'images',
            #                       reference_image_url = None,
            #                       custom_data = None,
            #                       backup_archive = None,
            #                       teamd_id = 1,
            #                       import_settings = {},
            #                   ),
            #                   ProjectInfo(id = 23,
            #                       name = 'lemons_test',
            #                       description = None,
            #                       size = '1177212',
            #                       readme = None,
            #                       workspace_id = 2,
            #                       images_count = None,
            #                       items_count = None,
            #                       datasets_count = None,
            #                       created_at = '2020-04-03T13:43:24.000Z',
            #                       updated_at = '2020-04-03T14:53:00.952Z',
            #                       type = 'images',
            #                       reference_image_url = None,
            #                       custom_data = None,
            #                       backup_archive = None),
            #                       teamd_id = 1,
            #                       import_settings = {},
            #                   )
            #                 ]
            # }

        """

        method = "projects.list.all"

        request_body = {}
        if filters is not None:
            request_body[ApiField.FILTER] = filters
        if skip_exported is not None:
            request_body[ApiField.SKIP_EXPORTED] = skip_exported
        if sort is not None:
            request_body[ApiField.SORT] = sort
        if sort_order is not None:
            request_body[ApiField.SORT_ORDER] = sort_order
        if per_page is not None:
            request_body[ApiField.PER_PAGE] = per_page
        if page is not None and page != "all":
            request_body[ApiField.PAGE] = page
        if account_type is not None:
            logger.warning(
                "The 'account_type' parameter is deprecated. The result will not be filtered by account type. To filter received ProjectInfos, you could use the 'team_id' from the ProjectInfo object to get TeamInfo and check the account type."
            )

        first_response = self._api.post(method, request_body).json()

        total = first_response.get(ApiField.TOTAL)
        per_page = first_response.get("perPage")
        pages_count = first_response.get("pagesCount")

        def _convert_entities(response_dict: dict):
            """
            Convert entities dict to ProjectInfo
            """
            response_dict[ApiField.ENTITIES] = [
                self._convert_json_info(item) for item in response_dict[ApiField.ENTITIES]
            ]

        if page == "all":
            results = first_response["entities"]
            if pages_count == 1 and len(results) == total:
                pass
            else:
                for page_idx in range(2, pages_count + 1):
                    temp_resp = self._api.post(
                        method, {**request_body, "page": page_idx, "per_page": per_page}
                    )
                    temp_items = temp_resp.json()["entities"]
                    results.extend(temp_items)

                if len(results) != total:
                    raise RuntimeError(
                        "Method {!r}: error during pagination, some items are missed".format(method)
                    )
            pass
        elif page is None or page <= pages_count:
            pass
        else:
            raise RuntimeError(
                f"Method {method}: error during pagination, some items are missed. Number of total pages [{pages_count}] is less than the page number requested [{page}]."
            )
        _convert_entities(first_response)
        return first_response



import argparse
import asyncio
import os
import signal
import sys
import time

import supervisely as sly
from supervisely.project.download import download_to_cache
from supervisely.project.project import _download_project, _download_project_async

LOG_LEVEL = "INFO"
# LOG_LEVEL = "DEBUG"
PROJECT_ID = 325865  #  41862
DATSET_ID = 98429
home_dir = os.path.expanduser("~")
common_path = os.path.join(home_dir, "test_project_download/")
save_path = os.path.join(common_path, "old/")
save_path_async = os.path.join(common_path, "async/")
sly.fs.ensure_base_path(common_path)
api = sly.Api.from_env()

# sly.fs.clean_dir(common_path)


def main_dpa(project_id: int, semaphore_size: int):
    if semaphore_size is None:
        sema = None
    else:
        sema = asyncio.Semaphore(semaphore_size)
    start = time.time()
    loop = asyncio.get_event_loop()
    loop.run_until_complete(
        _download_project_async(
            api, project_id, save_path_async, semaphore=sema, resume_download=True
        )
    )
    finish = time.time() - start
    print(f"Time taken for async method: {finish}")
    print(f"Project downloaded to {save_path_async}")


def main_dptc():
    start = time.time()
    download_to_cache(api, PROJECT_ID)
    finish = time.time() - start
    print(f"Time taken for cache method: {finish}")


def main_dps(project_id: int):
    start = time.time()
    _download_project(api, project_id, save_path)
    finish = time.time() - start
    print(f"Time taken for sync method: {finish}")


def compare_downloads(project_id: int, semaphore_size: int):
    main_dps(project_id)
    sly.fs.clean_dir(save_path)
    main_dpa(project_id, semaphore_size)
    sly.fs.clean_dir(save_path_async)


def ann_db():
    imgs = api.image.get_list(DATSET_ID)
    img_ids = [img.id for img in imgs]
    pbar = sly.tqdm_sly(desc="Downloading annotations", unit="B", unit_scale=True)
    loop = asyncio.get_event_loop()
    anns = loop.run_until_complete(
        api.annotation.download_batch_async(
            DATSET_ID,
            img_ids,
            semaphore=asyncio.Semaphore(200),
            progress_cb=pbar,
            progress_cb_type="size",
            # force_metadata_for_links=False,
        )
    )


if __name__ == "__main__":
    # parser = argparse.ArgumentParser(description="Download and save project as files (async)")
    # parser.add_argument(
    #     "--server", type=str, default=os.environ["SERVER_ADDRESS"], help="Server address"
    # )
    # parser.add_argument("--token", type=str, default=os.environ["API_TOKEN"], help="API token")
    # parser.add_argument("--id", type=int, default=PROJECT_ID, help="ID of the project to download")
    # parser.add_argument(
    #     "--semaphore", type=int, default=50, help="Semaphore size for async download"
    # )
    # args = parser.parse_args()
    try:
        # ann_db()
        # api = sly.Api(args.server, args.token)
        # api.logger.setLevel(LOG_LEVEL)
        main_dpa(PROJECT_ID, 200)  # to download and save project as files (async)
        # main_dps(args.id)  # to download and save project as files (sync)
        # compare_downloads(args.id, args.semaphore)  # to compare the time taken for downloading and saving project as files (sync vs async)
    except KeyboardInterrupt:
        sly.logger.info("Stopped by user")



from typing import Any, Dict

from supervisely import Api
from supervisely.app.widgets import (
    Button,
    Card,
    Container,
    DoneLabel,
    Empty,
    Field,
    FolderThumbnail,
    Input,
    ReportThumbnail,
    SelectCudaDevice,
    Text,
)


class TrainingProcess:
    title = "Training Process"
    description = "Manage training process"
    lock_message = "Select hyperparametrs to unlock"

    def __init__(self, app_options: Dict[str, Any]):
        self.display_widgets = []
        self.success_message_text = (
            "Training completed. Training artifacts were uploaded to Team Files. "
            "You can find and open tensorboard logs in the artifacts folder via the "
            "<a href='https://ecosystem.supervisely.com/apps/tensorboard-logs-viewer' target='_blank'>Tensorboard</a> app."
        )
        self.app_options = app_options

        # GUI Components
        # Optional Select CUDA device
        if self.app_options.get("device_selector", False):
            self.select_device = SelectCudaDevice()
            self.select_cuda_device_field = Field(
                title="Select CUDA device",
                description="The device on which the model will be trained",
                content=self.select_device,
            )
            self.display_widgets.extend([self.select_cuda_device_field])
        # -------------------------------- #

        self.experiment_name_input = Input("Enter experiment name")
        self.experiment_name_field = Field(
            title="Experiment name",
            description="Experiment name will be saved to experiment_info.json",
            content=self.experiment_name_input,
        )

        self.start_button = Button("Start")
        self.stop_button = Button("Stop", button_type="danger")
        self.stop_button.hide()  # @TODO: implement stop and hide stop button until training starts
        button_container = Container(
            [self.start_button, self.stop_button, Empty()],
            "horizontal",
            overflow="wrap",
            fractions=[1, 1, 10],
            gap=1,
        )

        self.validator_text = Text("")
        self.validator_text.hide()

        self.artifacts_thumbnail = FolderThumbnail()
        self.artifacts_thumbnail.hide()

        self.display_widgets.extend(
            [
                self.experiment_name_field,
                button_container,
                self.validator_text,
                self.artifacts_thumbnail,
            ]
        )
        # -------------------------------- #

        # Optional Model Benchmark
        if app_options.get("model_benchmark", False):
            self.model_benchmark_report_thumbnail = ReportThumbnail()
            self.model_benchmark_report_thumbnail.hide()
            self.display_widgets.extend([self.model_benchmark_report_thumbnail])
        # -------------------------------- #

        self.container = Container(self.display_widgets)
        self.card = Card(
            title=self.title,
            description=self.description,
            content=self.container,
            lock_message=self.lock_message,
        )
        self.card.lock()

    @property
    def widgets_to_disable(self) -> list:
        widgets = [self.experiment_name_input]
        if self.app_options.get("device_selector", False):
            widgets.append(self.experiment_name_input)
        return widgets

    def validate_step(self) -> bool:
        return True

    def get_device(self) -> str:
        if self.app_options.get("device_selector", False):
            return self.select_device.get_device()
        else:
            return "cuda:0"

    def get_experiment_name(self) -> str:
        return self.experiment_name_input.get_value()

    def set_experiment_name(self, experiment_name) -> None:
        self.experiment_name_input.set_value(experiment_name)



from supervisely.nn.benchmark.comparison.detection_visualization.vis_metrics.vis_metric import (
    BaseVisMetric,
)
from supervisely.nn.benchmark.visualization.widgets import (
    ChartWidget,
    CollapseWidget,
    MarkdownWidget,
    NotificationWidget,
    TableWidget,
)


class PrecisionRecallF1(BaseVisMetric):
    MARKDOWN = "markdown_PRF1"
    MARKDOWN_PRECISION_TITLE = "markdown_precision_per_class_title"
    MARKDOWN_RECALL_TITLE = "markdown_recall_per_class_title"
    MARKDOWN_F1_TITLE = "markdown_f1_per_class_title"

    @property
    def markdown_widget(self) -> MarkdownWidget:
        text: str = getattr(self.vis_texts, self.MARKDOWN).format(
            self.vis_texts.definitions.f1_score
        )
        return MarkdownWidget(name=self.MARKDOWN, title="Precision, Recall, F1-score", text=text)

    @property
    def precision_per_class_title_md(self) -> MarkdownWidget:
        text: str = getattr(self.vis_texts, self.MARKDOWN_PRECISION_TITLE)
        text += self.vis_texts.clickable_label
        return MarkdownWidget(
            name="markdown_precision_per_class", title="Precision by Class", text=text
        )

    @property
    def recall_per_class_title_md(self) -> MarkdownWidget:
        text: str = getattr(self.vis_texts, self.MARKDOWN_RECALL_TITLE)
        text += self.vis_texts.clickable_label
        return MarkdownWidget(name="markdown_recall_per_class", title="Recall by Class", text=text)

    @property
    def f1_per_class_title_md(self) -> MarkdownWidget:
        text: str = getattr(self.vis_texts, self.MARKDOWN_F1_TITLE)
        text += self.vis_texts.clickable_label
        return MarkdownWidget(name="markdown_f1_per_class", title="F1-score by Class", text=text)

    @property
    def chart_main_widget(self) -> ChartWidget:
        chart = ChartWidget(name="chart_PRF1", figure=self.get_main_figure())
        chart.set_click_data(
            gallery_id=self.explore_modal_table.id,
            click_data=self.get_click_data_main(),
            chart_click_extra="'getKey': (payload) => `${payload.points[0].curveNumber}`,",
        )
        return chart

    @property
    def chart_recall_per_class_widget(self) -> ChartWidget:
        chart = ChartWidget(
            name="chart_recall_per_class",
            figure=self.get_recall_per_class_figure(),
        )
        chart.set_click_data(
            gallery_id=self.explore_modal_table.id,
            click_data=self.get_per_class_click_data(),
            chart_click_extra="'getKey': (payload) => `${payload.points[0].curveNumber}${'_'}${payload.points[0].label}`,",
        )
        return chart

    @property
    def chart_precision_per_class_widget(self) -> ChartWidget:
        chart = ChartWidget(
            name="chart_precision_per_class",
            figure=self.get_precision_per_class_figure(),
        )
        chart.set_click_data(
            gallery_id=self.explore_modal_table.id,
            click_data=self.get_per_class_click_data(),
            chart_click_extra="'getKey': (payload) => `${payload.points[0].curveNumber}${'_'}${payload.points[0].label}`,",
        )
        return chart

    @property
    def chart_f1_per_class_widget(self) -> ChartWidget:
        chart = ChartWidget(name="chart_f1_per_class", figure=self.get_f1_per_class_figure())
        chart.set_click_data(
            gallery_id=self.explore_modal_table.id,
            click_data=self.get_per_class_click_data(),
            chart_click_extra="'getKey': (payload) => `${payload.points[0].curveNumber}${'_'}${payload.points[0].label}`,",
        )
        return chart

    @property
    def table_widget(self) -> TableWidget:
        res = {}

        columns = [" ", "Precision", "Recall", "F1-score"]
        res["content"] = []
        for i, eval_result in enumerate(self.eval_results, 1):
            precision = round(eval_result.mp.json_metrics()["precision"], 2)
            recall = round(eval_result.mp.json_metrics()["recall"], 2)
            f1 = round(eval_result.mp.json_metrics()["f1"], 2)
            model_name = f"[{i}] {eval_result.name}"
            row = [model_name, precision, recall, f1]
            dct = {
                "row": row,
                "id": model_name,
                "items": row,
            }
            res["content"].append(dct)

        columns_options = [
            {"disableSort": True},
            {"disableSort": True},
            {"disableSort": True},
            {"disableSort": True},
        ]

        res["columns"] = columns
        res["columnsOptions"] = columns_options

        return TableWidget(
            name="table_precision_recall_f1",
            data=res,
            show_header_controls=False,
            # main_column=columns[0],
            fix_columns=1,
        )

    def get_main_figure(self):  # -> Optional[go.Figure]:
        import plotly.graph_objects as go  # pylint: disable=import-error

        fig = go.Figure()

        classes_cnt = len(self.eval_results[0].mp.cat_names)
        for i, eval_result in enumerate(self.eval_results, 1):
            precision = eval_result.mp.json_metrics()["precision"]
            recall = eval_result.mp.json_metrics()["recall"]
            f1 = eval_result.mp.json_metrics()["f1"]
            model_name = f"[{i}] {eval_result.model_name}"
            fig.add_trace(
                go.Bar(
                    x=["Precision", "Recall", "F1-score"],
                    y=[precision, recall, f1],
                    name=model_name,
                    width=0.2 if classes_cnt >= 5 else None,
                    marker=dict(color=eval_result.color),
                )
            )

        fig.update_layout(
            barmode="group",
            xaxis_title="Metric",
            yaxis_title="Value",
            yaxis=dict(range=[0, 1.1]),
            width=700 if classes_cnt < 5 else None,
        )

        return fig

    def get_recall_per_class_figure(self):
        import plotly.graph_objects as go  # pylint: disable=import-error

        fig = go.Figure()
        classes_cnt = len(self.eval_results[0].mp.cat_names)
        for i, eval_result in enumerate(self.eval_results, 1):
            model_name = f"[{i}] {eval_result.model_name}"
            sorted_by_f1 = eval_result.mp.per_class_metrics().sort_values(by="f1")

            fig.add_trace(
                go.Bar(
                    y=sorted_by_f1["recall"],
                    x=sorted_by_f1["category"],
                    name=f"{model_name} Recall",
                    width=0.2 if classes_cnt < 5 else None,
                    marker=dict(color=eval_result.color),
                )
            )

        fig.update_layout(
            barmode="group",
            bargap=0.15,
            bargroupgap=0.05,
            width=700 if classes_cnt < 5 else None,
        )
        fig.update_xaxes(title_text="Class")
        fig.update_yaxes(title_text="Recall", range=[0, 1])
        return fig

    def get_per_class_click_data(self):
        res = {}
        res["layoutTemplate"] = [None, None, None]
        res["clickData"] = {}
        for i, eval_result in enumerate(self.eval_results):
            model_name = f"Model [{i + 1}] {eval_result.model_name}"
            for key, v in eval_result.click_data.objects_by_class.items():
                click_data = res["clickData"].setdefault(f"{i}_{key}", {})
                img_ids, obj_ids = set(), set()
                title = f"{model_name}. Class {key}: {len(v)} object{'s' if len(v) > 1 else ''}"
                click_data["title"] = title

                for x in v:
                    img_ids.add(x["dt_img_id"])
                    obj_ids.add(x["dt_obj_id"])

                click_data["imagesIds"] = list(img_ids)
                click_data["filters"] = [
                    {
                        "type": "tag",
                        "tagId": "confidence",
                        "value": [eval_result.f1_optimal_conf, 1],
                    },
                    {"type": "tag", "tagId": "outcome", "value": "TP"},
                    {"type": "specific_objects", "tagId": None, "value": list(obj_ids)},
                ]
        return res

    def get_precision_per_class_figure(self):
        import plotly.graph_objects as go  # pylint: disable=import-error

        fig = go.Figure()
        classes_cnt = len(self.eval_results[0].mp.cat_names)
        for i, eval_result in enumerate(self.eval_results, 1):
            model_name = f"[{i}] {eval_result.model_name}"
            sorted_by_f1 = eval_result.mp.per_class_metrics().sort_values(by="f1")

            fig.add_trace(
                go.Bar(
                    y=sorted_by_f1["precision"],
                    x=sorted_by_f1["category"],
                    name=f"{model_name} Precision",
                    width=0.2 if classes_cnt < 5 else None,
                    marker=dict(color=eval_result.color),
                )
            )

        fig.update_layout(
            barmode="group",
            bargap=0.15,
            bargroupgap=0.05,
            width=700 if classes_cnt < 5 else None,
        )
        fig.update_xaxes(title_text="Class")
        fig.update_yaxes(title_text="Precision", range=[0, 1])
        return fig

    def get_f1_per_class_figure(self):
        import plotly.graph_objects as go  # pylint: disable=import-error

        fig = go.Figure()
        classes_cnt = len(self.eval_results[0].mp.cat_names)
        for i, eval_result in enumerate(self.eval_results, 1):
            model_name = f"[{i}] {eval_result.model_name}"
            sorted_by_f1 = eval_result.mp.per_class_metrics().sort_values(by="f1")

            fig.add_trace(
                go.Bar(
                    y=sorted_by_f1["f1"],
                    x=sorted_by_f1["category"],
                    name=f"{model_name} F1-score",
                    width=0.2 if classes_cnt < 5 else None,
                    marker=dict(color=eval_result.color),
                )
            )

        fig.update_layout(
            barmode="group",
            bargap=0.15,
            bargroupgap=0.05,
            width=700 if classes_cnt < 5 else None,
        )
        fig.update_xaxes(title_text="Class")
        fig.update_yaxes(title_text="F1-score", range=[0, 1])
        return fig

    def get_click_data_main(self):
        res = {}
        res["layoutTemplate"] = [None, None, None]
        res["clickData"] = {}

        for i, eval_result in enumerate(self.eval_results):
            model_name = f"Model [{i + 1}] {eval_result.model_name}"
            click_data = res["clickData"].setdefault(i, {})
            img_ids, obj_ids = set(), set()
            objects_cnt = 0
            for outcome, matched_obj in eval_result.click_data.outcome_counts.items():
                if outcome == "TP":  # TODO: check if this is correct
                    objects_cnt += len(matched_obj)
                    for x in matched_obj:
                        img_ids.add(x["dt_img_id"])
                        obj_ids.add(x["dt_obj_id"])

            click_data["title"] = f"{model_name}, {objects_cnt} objects"
            click_data["imagesIds"] = list(img_ids)
            click_data["filters"] = [
                {"type": "tag", "tagId": "confidence", "value": [eval_result.f1_optimal_conf, 1]},
                {"type": "tag", "tagId": "outcome", "value": "TP"},
                {"type": "specific_objects", "tagId": None, "value": list(obj_ids)},
            ]

        return res



# coding: utf-8

import asyncio
import time
import traceback

import httpx
import requests

CONNECTION_ERROR = "Temporary connection error, please wait ..."
AGENT_CONNECTION_ERROR = "Temporary connection error (agent ping), please wait ..."

REQUEST_FAILED = "Request has failed. This may be due to connection problems or invalid requests. "
SPECIAL_RECONNECT_ERROR = (
    "Agent should call AgentConnected or AgentPing before attempting any other request"
)
RETRY_STATUS_CODES = {
    408,  # Request Timeout
    429,  # Too Many Requests
    499,  # Client Closed Request (Nginx)
    500,  # Internal Server Error
    502,  # Bad Gateway
    503,  # Service Unavailable
    504,  # Gateway Timeout
    509,  # Bandwidth Limit Exceeded (Apache)
    598,  # Network read timeout error
    599,  # Network connect timeout error
}


async def process_requests_exception_async(
    external_logger,
    exc,
    api_method_name,
    url,
    verbose=True,
    swallow_exc=False,
    sleep_sec=None,
    response=None,
    retry_info=None,
):
    recommended_sleep = None
    try:
        if hasattr(exc, "response") and exc.response.status_code == 429:
            recommended_sleep = exc.response.headers.get("Retry-After")
        elif response is not None and response.status_code == 429:
            recommended_sleep = response.headers.get("Retry-After")
        if recommended_sleep and int(recommended_sleep) > sleep_sec:
            sleep_sec = int(recommended_sleep)
    except Exception:
        pass

    is_connection_error = isinstance(
        exc,
        (
            requests.exceptions.ConnectionError,
            requests.exceptions.Timeout,
            requests.exceptions.TooManyRedirects,
            requests.exceptions.ChunkedEncodingError,
            httpx.NetworkError,
            httpx.TimeoutException,
            httpx.TooManyRedirects,
            httpx.ProtocolError,
        ),
    )

    is_server_retryable_error = (
        isinstance(exc, (requests.exceptions.HTTPError, httpx.HTTPStatusError))
        and hasattr(exc, "response")
        and (exc.response.status_code in RETRY_STATUS_CODES)
    )

    is_need_ping_error = False
    if (
        isinstance(exc, (requests.exceptions.HTTPError, httpx.HTTPStatusError))
        and hasattr(exc, "response")
        and (exc.response.status_code == 400)
    ):
        try:
            server_explanation = exc.response.json()
            is_need_ping_error = server_explanation.get("error", None) == SPECIAL_RECONNECT_ERROR
        except (AttributeError, ValueError):
            pass

    if is_connection_error or is_server_retryable_error:
        await process_retryable_request_async(
            external_logger,
            exc,
            api_method_name,
            url,
            CONNECTION_ERROR,
            verbose=verbose,
            swallow_exc=swallow_exc,
            sleep_sec=sleep_sec,
            retry_info=retry_info,
        )
    elif is_need_ping_error:
        await process_retryable_request_async(
            external_logger,
            exc,
            api_method_name,
            url,
            AGENT_CONNECTION_ERROR,
            verbose=verbose,
            swallow_exc=swallow_exc,
            sleep_sec=sleep_sec,
            retry_info=retry_info,
        )
    elif response is None:
        process_unhandled_request(external_logger, exc)
    elif isinstance(exc, (requests.exceptions.HTTPError, httpx.HTTPStatusError)):
        process_invalid_request(external_logger, exc, response, verbose)
    else:
        process_unhandled_request(external_logger, exc)


def process_requests_exception(
    external_logger,
    exc,
    api_method_name,
    url,
    verbose=True,
    swallow_exc=False,
    sleep_sec=None,
    response=None,
    retry_info=None,
):
    recommended_sleep = None
    try:
        if hasattr(exc, "response") and exc.response.status_code == 429:
            recommended_sleep = exc.response.headers.get("Retry-After")
        elif response is not None and response.status_code == 429:
            recommended_sleep = response.headers.get("Retry-After")
        if recommended_sleep and int(recommended_sleep) > sleep_sec:
            sleep_sec = int(recommended_sleep)
    except Exception:
        pass

    is_connection_error = isinstance(
        exc,
        (
            requests.exceptions.ConnectionError,
            requests.exceptions.Timeout,
            requests.exceptions.TooManyRedirects,
            requests.exceptions.ChunkedEncodingError,
            httpx.NetworkError,
            httpx.TimeoutException,
            httpx.TooManyRedirects,
            httpx.ProtocolError,
        ),
    )

    is_server_retryable_error = (
        isinstance(exc, (requests.exceptions.HTTPError, httpx.HTTPStatusError))
        and hasattr(exc, "response")
        and (exc.response.status_code in RETRY_STATUS_CODES)
    )

    is_need_ping_error = False
    if (
        isinstance(exc, (requests.exceptions.HTTPError, httpx.HTTPStatusError))
        and hasattr(exc, "response")
        and (exc.response.status_code == 400)
    ):
        try:
            server_explanation = exc.response.json()
            is_need_ping_error = server_explanation.get("error", None) == SPECIAL_RECONNECT_ERROR
        except (AttributeError, ValueError):
            pass

    if is_connection_error or is_server_retryable_error:
        process_retryable_request(
            external_logger,
            exc,
            api_method_name,
            url,
            CONNECTION_ERROR,
            verbose=verbose,
            swallow_exc=swallow_exc,
            sleep_sec=sleep_sec,
            retry_info=retry_info,
        )
    elif is_need_ping_error:
        process_retryable_request(
            external_logger,
            exc,
            api_method_name,
            url,
            AGENT_CONNECTION_ERROR,
            verbose=verbose,
            swallow_exc=swallow_exc,
            sleep_sec=sleep_sec,
            retry_info=retry_info,
        )
    elif response is None:
        process_unhandled_request(external_logger, exc)
    elif isinstance(exc, (requests.exceptions.HTTPError, httpx.HTTPStatusError)):
        process_invalid_request(external_logger, exc, response, verbose)
    else:
        process_unhandled_request(external_logger, exc)


async def process_retryable_request_async(
    external_logger,
    exc,
    api_method_name,
    url,
    user_message,
    verbose=True,
    swallow_exc=False,
    sleep_sec=None,
    retry_info=None,
):
    if retry_info is not None:
        retry_idx = retry_info["retry_idx"]
        retry_limit = retry_info["retry_limit"]
        user_message = "{}:  Retrying ({}/{}).".format(user_message, retry_idx, retry_limit)
    if verbose:
        external_logger.warn(
            user_message, extra={"method": api_method_name, "url": url, "details": str(exc)}
        )

    if sleep_sec is not None:
        await asyncio.sleep(sleep_sec)

    if not swallow_exc:
        raise exc


def process_retryable_request(
    external_logger,
    exc,
    api_method_name,
    url,
    user_message,
    verbose=True,
    swallow_exc=False,
    sleep_sec=None,
    retry_info=None,
):
    if retry_info is not None:
        retry_idx = retry_info["retry_idx"]
        retry_limit = retry_info["retry_limit"]
        user_message = "{}:  Retrying ({}/{}).".format(user_message, retry_idx, retry_limit)
    if verbose:
        external_logger.warn(
            user_message, extra={"method": api_method_name, "url": url, "details": str(exc)}
        )

    if sleep_sec is not None:
        time.sleep(sleep_sec)

    if not swallow_exc:
        raise exc


def process_invalid_request(external_logger, exc, response, verbose=True):
    if type(response) in (httpx.Response, requests.Response):
        reason = (
            response.content.decode("utf-8")
            if not hasattr(response, "is_stream_consumed")
            else "Content is not acessible for streaming responses"
        )
        status_code = response.status_code
        url = response.url
    else:
        reason = "Reason is unknown"
        status_code = None
        url = None
    if verbose:
        external_logger.warn(
            REQUEST_FAILED,
            extra={
                "reason": response.content.decode("utf-8"),
                "status_code": response.status_code,
                "url": response.url,
            },
        )
    raise exc


def process_unhandled_request(external_logger, exc):
    exc_str = str(exc)
    external_logger.error(traceback.format_exc(), exc_info=True, extra={"exc_str": exc_str})
    raise RuntimeError(REQUEST_FAILED + "Last failure: {!r}".format(exc_str))



# isort: skip_file
from supervisely.nn.benchmark.utils.detection.calculate_metrics import calculate_metrics
from supervisely.nn.benchmark.utils.detection.metric_provider import MetricProvider
from supervisely.nn.benchmark.utils.detection.sly2coco import sly2coco
from supervisely.nn.benchmark.utils.detection.utlis import read_coco_datasets
from supervisely.nn.benchmark.utils.detection.utlis import try_set_conf_auto

from supervisely.nn.benchmark.utils.semantic_segmentation.calculate_metrics import (
    calculate_metrics as calculate_semsegm_metrics,
)
from supervisely.nn.benchmark.utils.semantic_segmentation.evaluator import Evaluator
from supervisely.nn.benchmark.utils.semantic_segmentation.loader import build_segmentation_loader



# coding: utf-8

from typing import Any, Dict, Optional

from supervisely.api.module_api import ApiField, ModuleApiBase
from supervisely.collection.str_enum import StrEnum


class VideoAnnotationToolAction(StrEnum):
    JOBS_DISABLE_CONTROLS = "jobs/disableControls"
    """"""
    JOBS_ENABLE_CONTROLS = "jobs/enableControls"
    """"""
    JOBS_DISABLE_SUBMIT = "jobs/disableSubmit"
    """"""
    JOBS_ENABLE_SUBMIT = "jobs/enableSubmit"
    """"""
    JOBS_DISABLE_CONFIRM = "jobs/disableConfirm"
    """"""
    JOBS_ENABLE_CONFIRM = "jobs/enableConfirm"
    """"""
    ENTITIES_SET_INTITY = "entities/setEntity"
    """"""


class VideoAnnotationToolApi(ModuleApiBase):
    def disable_job_controls(self, session_id: str) -> Dict[str, Any]:
        """Disables controls of the labeling jobs. Buttons: Sumbit job, Confirm video.

        :param session_id: ID of the session in the Video Labeling Tool which controls should be disabled.
        :type session_id: str
        :return: Response from API server in JSON format.
        :rtype: Dict[str, Any]
        """
        return self._act(
            session_id,
            VideoAnnotationToolAction.JOBS_DISABLE_CONTROLS,
            {},
        )

    def enable_job_controls(self, session_id: str) -> Dict[str, Any]:
        """Enables controls of the labeling jobs. Buttons: Sumbit job, Confirm video.

        :param session_id: ID of the session in the Video Labeling Tool which controls should be enabled.
        :type session_id: str
        :return: Response from API server in JSON format.
        :rtype: Dict[str, Any]
        """
        return self._act(
            session_id,
            VideoAnnotationToolAction.JOBS_ENABLE_CONTROLS,
            {},
        )
        
    def disable_submit_button(self, session_id: str) -> Dict[str, Any]:
        """Disables submit button of the labeling jobs.

        :param session_id: ID of the session in the Video Labeling Tool which submit button should be disabled.
        :type session_id: str
        :return: Response from API server in JSON format.
        :rtype: Dict[str, Any]
        """
        return self._act(
            session_id,
            VideoAnnotationToolAction.JOBS_DISABLE_SUBMIT,
            {},
        )
        
    def enable_submit_button(self, session_id: str) -> Dict[str, Any]:
        """Enables submit button of the labeling jobs.

        :param session_id: ID of the session in the Video Labeling Tool which submit button should be enabled.
        :type session_id: str
        :return: Response from API server in JSON format.
        :rtype: Dict[str, Any]
        """
        return self._act(
            session_id,
            VideoAnnotationToolAction.JOBS_ENABLE_SUBMIT,
            {},
        )
        
    def disable_confirm_button(self, session_id: str) -> Dict[str, Any]:
        """Disables confirm button of the labeling jobs.

        :param session_id: ID of the session in the Video Labeling Tool which confirm button should be disabled.
        :type session_id: str
        :return: Response from API server in JSON format.
        :rtype: Dict[str, Any]
        """
        return self._act(
            session_id,
            VideoAnnotationToolAction.JOBS_DISABLE_CONFIRM,
            {},
        )
    
    def enable_confirm_button(self, session_id: str) -> Dict[str, Any]:
        """Enables confirm button of the labeling jobs.
        
        :param session_id: ID of the session in the Video Labeling Tool which confirm button should be enabled.
        :type session_id: str
        :return: Response from API server in JSON format.
        :rtype: Dict[str, Any]
        """
        return self._act(
            session_id,
            VideoAnnotationToolAction.JOBS_ENABLE_CONFIRM,
            {},
        )

    def set_video(self, session_id: str, video_id: int, frame: Optional[int] = 0) -> Dict[str, Any]:
        """Sets video in the Video Labeling Tool and switches to the specified frame
        if frame number is provided.
        NOTE: Video from the same dataset should be set in the Video Labeling Tool.

        :param session_id: ID of the session in the Video Labeling Tool where video should be set.
        :type session_id: str
        :param video_id: ID of the video in the same dataset which should be set.
        :type video_id: int
        :param frame: Frame number which should be set, defaults to 0.
        :type frame: Optional[int]
        :return: Response from API server in JSON format.
        :rtype: Dict[str, Any]
        """

        return self._act(
            session_id,
            VideoAnnotationToolAction.ENTITIES_SET_INTITY,
            {
                ApiField.ENTITY_ID: video_id,
                ApiField.FRAME: frame,
            },
        )

    def _act(self, session_id: int, action: VideoAnnotationToolAction, payload: dict):
        data = {
            ApiField.SESSION_ID: session_id,
            ApiField.ACTION: str(action),
            ApiField.PAYLOAD: payload,
        }
        resp = self._api.post("/annotation-tool.run-action", data)

        return resp.json()



from __future__ import annotations

import os
import pickle
from typing import Dict, List, Optional, Union

import yaml

from supervisely.app.widgets import SlyTqdm
from supervisely.task.progress import tqdm_sly


class BaseEvalResult:
    def __init__(self, directory: Optional[str] = None):
        self.directory = directory
        self.inference_info: Dict = None
        self.speedtest_info: Dict = None
        self.eval_data: Dict = None
        self.mp = None

        if self.directory is not None:
            self._read_files(self.directory)
            self._prepare_data()

    @classmethod
    def from_evaluator(cls, evaulator: BaseEvaluator) -> BaseEvalResult:
        """Method to customize loading of the evaluation result."""
        raise NotImplementedError()

    @property
    def cv_task(self):
        return self.inference_info.get("task_type")

    @property
    def name(self) -> Union[str, None]:
        deploy_params = self.inference_info.get("deploy_params", {})
        return (
            deploy_params.get("checkpoint_name")
            or deploy_params.get("model_name")
            or self.inference_info.get("model_name")
        )

    @property
    def gt_project_id(self) -> int:
        return self.inference_info.get("gt_project_id")

    @property
    def gt_dataset_ids(self) -> List[int]:
        return self.inference_info.get("gt_dataset_ids", None)

    @property
    def dt_project_id(self):
        return self.inference_info.get("dt_project_id")

    @property
    def pred_project_id(self):
        return self.dt_project_id

    @property
    def train_info(self):
        return self.inference_info.get("train_info", None)  # TODO: check

    @property
    def evaluator_app_info(self):
        return self.inference_info.get("evaluator_app_info", None)  # TODO: check

    @property
    def val_images_cnt(self):
        return self.inference_info.get("val_images_cnt", None)  # TODO: check

    @property
    def classes_whitelist(self):
        return self.inference_info.get("inference_settings", {}).get("classes", [])  # TODO: check

    def _read_files(self, path: str) -> None:
        """Read all necessary files from the directory"""
        raise NotImplementedError()

    def _prepare_data(self) -> None:
        """Prepare data to allow easy access to the data"""
        raise NotImplementedError()
    
    @property
    def key_metrics(self):
        raise NotImplementedError()


class BaseEvaluator:
    EVALUATION_PARAMS_YAML_PATH: str = None
    eval_result_cls = BaseEvalResult

    def __init__(
        self,
        gt_project_path: str,
        pred_project_path: str,
        result_dir: str = "./evaluation",
        progress: Optional[SlyTqdm] = None,
        items_count: Optional[int] = None,  # TODO: is it needed?
        classes_whitelist: Optional[List[str]] = None,
        evaluation_params: Optional[dict] = None,
    ):
        self.gt_project_path = gt_project_path
        self.pred_project_path = pred_project_path
        self.result_dir = result_dir
        self.total_items = items_count
        self.pbar = progress or tqdm_sly
        os.makedirs(result_dir, exist_ok=True)
        self.classes_whitelist = classes_whitelist

        if evaluation_params is None:
            evaluation_params = self._get_default_evaluation_params()
        self.evaluation_params = evaluation_params
        if self.evaluation_params:
            self.validate_evaluation_params(self.evaluation_params)

    def evaluate(self):
        raise NotImplementedError()

    @classmethod
    def load_yaml_evaluation_params(cls) -> Union[str, None]:
        if cls.EVALUATION_PARAMS_YAML_PATH is None:
            return None
        with open(cls.EVALUATION_PARAMS_YAML_PATH, "r") as f:
            return f.read()

    @classmethod
    def validate_evaluation_params(cls, evaluation_params: dict) -> None:
        pass

    @classmethod
    def _get_default_evaluation_params(cls) -> dict:
        if cls.EVALUATION_PARAMS_YAML_PATH is None:
            return {}
        else:
            return yaml.safe_load(cls.load_yaml_evaluation_params())

    def _dump_pickle(self, data, file_path):
        with open(file_path, "wb") as f:
            pickle.dump(data, f)

    def get_eval_result(self) -> BaseEvalResult:
        return self.eval_result_cls(self.result_dir)



import random
from pathlib import Path
from typing import List, Optional

from supervisely.api.api import Api
from supervisely.app.widgets import SlyTqdm
from supervisely.imaging.color import get_predefined_colors, rgb2hex
from supervisely.nn.benchmark.comparison.detection_visualization.visualizer import (
    DetectionComparisonVisualizer,
)
from supervisely.nn.benchmark.visualization.evaluation_result import EvalResult
from supervisely.task.progress import tqdm_sly


class ModelComparison:

    def __init__(
        self,
        api: Api,
        remote_eval_dirs: List[str],
        progress: Optional[SlyTqdm] = None,
        workdir: Optional[str] = "./benchmark/comparison",
    ):
        self.api = api
        self.progress = progress or tqdm_sly
        self.workdir = workdir
        self.remote_eval_dirs = remote_eval_dirs
        self.evaluation_results: List[EvalResult] = []

        colors = get_predefined_colors(len(remote_eval_dirs) * 5)  # for better visualizations
        random.shuffle(colors)
        for i, eval_dir in enumerate(remote_eval_dirs):
            local_path = str(Path(self.workdir, "eval_data"))
            eval_result = EvalResult(eval_dir, local_path, self.api, self.progress)
            self.evaluation_results.append(eval_result)
            eval_result.color = rgb2hex(colors[i])

        self.task_type = self.evaluation_results[0].inference_info.get("task_type")
        self._validate_eval_data()

        self.visualizer: DetectionComparisonVisualizer = None
        self.remote_dir = None

    def _validate_eval_data(self):
        """
        Validate the evaluation data before running the comparison.
        Make sure the benchmarks are done on the same project and datasets.
        """
        task_type = None
        img_names = None
        cat_names = None
        for eval_result in self.evaluation_results:
            next_task_type = eval_result.cv_task
            if not task_type is None:
                assert task_type == next_task_type, "Task types are different in the evaluations."
            task_type = next_task_type
            next_img_names = set(
                [img.get("file_name") for img in eval_result.coco_gt.imgs.values()]
            )
            if not img_names is None:
                assert img_names == next_img_names, "Images are different in the evaluations."
            img_names = next_img_names
            next_cat_names = set([cat.get("name") for cat in eval_result.coco_gt.cats.values()])
            if not cat_names is None:
                assert cat_names == next_cat_names, "Categories are different in the evaluations."
            cat_names = next_cat_names

    def get_metrics(self):
        pass

    def visualize(self):
        if self.visualizer is None:
            self.visualizer = DetectionComparisonVisualizer(self)
        self.visualizer.visualize()

    def upload_results(self, team_id: int, remote_dir: str, progress=None) -> str:
        self.remote_dir = self.visualizer.upload_results(team_id, remote_dir, progress)
        return self.remote_dir

    def get_report_link(self) -> str:
        if self.remote_dir is None:
            raise ValueError("Results are not uploaded yet.")
        report_link = self.remote_dir.rstrip("/") + "/template.vue"
        return report_link



from typing import Dict

from supervisely.nn.benchmark.semantic_segmentation.base_vis_metric import (
    SemanticSegmVisMetric,
)
from supervisely.nn.benchmark.visualization.widgets import GalleryWidget, MarkdownWidget


class ExplorePredictions(SemanticSegmVisMetric):
    MARKDOWN = "explore_predictions"
    GALLERY = "explore_predictions"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.clickable = True

    @property
    def md(self) -> MarkdownWidget:
        text = self.vis_texts.markdown_explorer
        return MarkdownWidget(self.MARKDOWN, "Explore Predictions", text)

    def gallery(self, opacity) -> GalleryWidget:
        gallery = GalleryWidget(self.GALLERY, columns_number=3, opacity=opacity)
        gallery.set_project_meta(self.eval_result.filtered_project_meta)
        gallery.add_image_left_header("Compare with GT")

        gallery.set_images(
            image_infos=self.eval_result.sample_images,
            ann_infos=self.eval_result.sample_anns,
        )
        gallery._gallery._update_filters()

        # set click data for diff gallery
        self.explore_modal_table.set_click_data(
            self.diff_modal_table.id,
            self.get_click_data(),
            get_key="(payload) => `${payload.annotation.image_id || payload.annotation.imageId}`",
        )

        gallery.set_click_data(
            self.diff_modal_table.id,
            self.get_click_data(),
            get_key="(payload) => `${payload.annotation.image_id || payload.annotation.imageId}`",
        )

        # set click data for explore gallery
        gallery.set_show_all_data(
            self.explore_modal_table.id,
            self.get_all_data(),
        )
        return gallery

    def get_all_data(self) -> dict:
        res = {}

        res["layoutTemplate"] = [None, None, None]
        click_data = res.setdefault("clickData", {})
        explore = click_data.setdefault("explore", {})
        explore["title"] = "Explore all predictions"
        images_ids = [d.pred_image_info.id for d in self.eval_result.matched_pair_data.values()]
        explore["imagesIds"] = images_ids

        return res

    def get_click_data(self) -> Dict:
        res = {}

        res["layoutTemplate"] = [
            {"columnTitle": "Original Image"},
            {"columnTitle": "Ground Truth Masks"},
            {"columnTitle": "Predicted Masks"},
        ]

        click_data = res.setdefault("clickData", {})

        for pairs_data in self.eval_result.matched_pair_data.values():
            gt = pairs_data.gt_image_info
            pred = pairs_data.pred_image_info
            diff = pairs_data.diff_image_info
            assert gt.name == pred.name == diff.name
            key = click_data.setdefault(str(pred.id), {})
            key["imagesIds"] = [diff.id, gt.id, pred.id]
            key["title"] = f"Image: {pred.name}"
        return res



# coding: utf-8

import asyncio
import base64
import copy
import hashlib
import json
import os
import random
import re
import string
import time
import urllib
from datetime import datetime
from functools import wraps
from tempfile import gettempdir
from typing import Any, Dict, List, Literal, Optional, Tuple

import numpy as np
import requests
from requests.utils import DEFAULT_CA_BUNDLE_PATH

from supervisely.io import env as sly_env
from supervisely.io import fs as sly_fs
from supervisely.sly_logger import logger

random.seed(time.time())


def rand_str(length):
    chars = string.ascii_letters + string.digits  # [A-z][0-9]
    return "".join((random.choice(chars)) for _ in range(length))


def generate_free_name(used_names, possible_name, with_ext=False, extend_used_names=False):
    res_name = possible_name
    new_suffix = 1
    while res_name in set(used_names):
        if with_ext is True:
            res_name = "{}_{:02d}{}".format(
                sly_fs.get_file_name(possible_name),
                new_suffix,
                sly_fs.get_file_ext(possible_name),
            )
        else:
            res_name = "{}_{:02d}".format(possible_name, new_suffix)
        new_suffix += 1
    if extend_used_names:
        used_names.add(res_name)
    return res_name


def generate_names(base_name, count):
    name = sly_fs.get_file_name(base_name)
    ext = sly_fs.get_file_ext(base_name)

    names = [base_name]
    for idx in range(1, count):
        names.append("{}_{:02d}{}".format(name, idx, ext))

    return names


def camel_to_snake(name):
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def snake_to_human(snake_str: str) -> str:
    """Return a human-readable string from a snake_case string.
    E.g. 'hello_world' -> 'Hello World'

    :param snake_str: snake_case string
    :type snake_str: str
    :return: Human-readable string
    :rtype: str
    """
    components = snake_str.split("_")
    return " ".join(word.capitalize() for word in components)


def take_with_default(v, default):
    return v if v is not None else default


def batched(seq, batch_size=50):
    for i in range(0, len(seq), batch_size):
        yield seq[i : i + batch_size]


def get_bytes_hash(bytes):
    return base64.b64encode(hashlib.sha256(bytes).digest()).decode("utf-8")


def get_string_hash(data):
    return base64.b64encode(hashlib.sha256(str.encode(data)).digest()).decode("utf-8")


def unwrap_if_numpy(x):
    return x.item() if isinstance(x, np.number) else x


def _dprint(json_data):
    print(json.dumps(json_data))


class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NpEncoder, self).default(obj)


COMMUNITY = "community"
ENTERPRISE = "enterprise"


def validate_percent(value):
    if 0 <= value <= 100:
        pass
    else:
        raise ValueError("Percent has to be in range [0; 100]")


def sizeof_fmt(num, suffix="B"):
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return "%3.1f %s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f %s%s" % (num, "Yi", suffix)


def _remove_sensitive_information(d: dict):
    new_dict = copy.deepcopy(d)
    fields = ["api_token", "API_TOKEN", "AGENT_TOKEN", "apiToken", "spawnApiToken"]
    for field in fields:
        if field in new_dict:
            new_dict[field] = "***"

    for parent_key in ["state", "context"]:
        if parent_key in new_dict and type(new_dict[parent_key]) is dict:
            for field in fields:
                if field in new_dict[parent_key]:
                    new_dict[parent_key][field] = "***"
    return new_dict


def validate_img_size(img_size):
    if not isinstance(img_size, (tuple, list)):
        raise TypeError(
            '{!r} has to be a tuple or a list. Given type "{}".'.format("img_size", type(img_size))
        )
    return tuple(img_size)


def is_development() -> bool:
    mode = os.environ.get("ENV", "development")
    if mode == "production":
        return False
    else:
        return True


def is_debug_with_sly_net() -> bool:
    mode = os.environ.get("DEBUG_WITH_SLY_NET")
    if mode is not None:
        return True
    else:
        return False


def is_docker():
    path = "/proc/self/cgroup"
    return (
        os.path.exists("/.dockerenv")
        or os.path.isfile(path)
        and any("docker" in line for line in open(path))
    )


def is_production() -> bool:
    return not is_development()


def is_community() -> bool:
    server_address = sly_env.server_address()

    if (
        server_address.rstrip("/") == "https://app.supervise.ly"
        or server_address.rstrip("/") == "https://app.supervisely.com"
    ):
        return True
    else:
        return False


def abs_url(relative_url: str) -> str:
    from supervisely.api.api import SERVER_ADDRESS

    server_address = os.environ.get(SERVER_ADDRESS, "")
    if server_address == "":
        logger.warn("SERVER_ADDRESS env variable is not defined")
    return urllib.parse.urljoin(server_address, relative_url)


def compress_image_url(
    url: str,
    width: Optional[int] = None,
    height: Optional[int] = None,
    quality: Optional[int] = 70,
) -> str:
    """NOTE: This function is deprecated. Use resize_image_url instead.
    Returns a URL to a compressed image with given parameters.

    :param url: Full Image storage URL, can be obtained from ImageInfo.
    :type url: str
    :param width: Width of the compressed image.
    :type width: int, optional
    :param height: Height of the compressed image.
    :type height: int, optional
    :param quality: Quality of the compressed image.
    :type quality: int, optional
    :return: Full URL to a compressed image.
    :rtype: str
    """
    if width is None:
        width = ""
    if height is None:
        height = ""
    return url.replace(
        "/image-converter",
        f"/previews/{width}x{height},jpeg,q{quality}/image-converter",
    )


def resize_image_url(
    full_storage_url: str,
    ext: Literal["jpeg", "png"] = "jpeg",
    method: Literal["fit", "fill", "fill-down", "force", "auto"] = "auto",
    width: int = 0,
    height: int = 0,
    quality: int = 70,
) -> str:
    """Returns a URL to a resized image with given parameters.
    Default sizes are 0, which means that the image will not be resized,
    just compressed if the extension is jpeg to the given quality.
    Learn more about resize parameters `here <https://docs.imgproxy.net/usage/processing#resize>`_.

    :param full_storage_url: Full Image storage URL, can be obtained from ImageInfo.
    :type full_storage_url: str
    :param ext: Image extension, jpeg or png.
    :type ext: Literal["jpeg", "png"], optional
    :param method: Resize type, fit, fill, fill-down, force, auto.
    :type method: Literal["fit", "fill", "fill-down", "force", "auto"], optional
    :param width: Width of the resized image.
    :type width: int, optional
    :param height: Height of the resized image.
    :type height: int, optional
    :param quality: Quality of the resized image.
    :type quality: int, optional
    :return: Full URL to a resized image.
    :rtype: str

    :Usage example:

    .. code-block:: python

        import supervisely as sly
        from supervisely_utils import resize_image_url

        api = sly.Api(server_address, token)

        image_id = 376729
        img_info = api.image.get_info_by_id(image_id)

        img_resized_url = resize_image_url(
            img_info.full_storage_url, ext="jpeg", method="fill", width=512, height=256)
        print(img_resized_url)
        # Output: https://app.supervisely.com/previews/q/ext:jpeg/resize:fill:512:256:0/q:70/plain/h5un6l2bnaz1vj8a9qgms4-public/images/original/2/X/Re/<image_name>.jpg
    """
    # original url example: https://app.supervisely.com/h5un6l2bnaz1vj8a9qgms4-public/images/original/2/X/Re/<image_name>.jpg
    # resized url example:  https://app.supervisely.com/previews/q/ext:jpeg/resize:fill:300:0:0/q:70/plain/h5un6l2bnaz1vj8a9qgms4-public/images/original/2/X/Re/<image_name>.jpg
    # to add: previews/q/ext:jpeg/resize:fill:300:0:0/q:70/plain/
    try:
        parsed_url = urllib.parse.urlparse(full_storage_url)
        server_address = f"{parsed_url.scheme}://{parsed_url.netloc}"

        resize_string = f"previews/q/ext:{ext}/resize:{method}:{width}:{height}:0/q:{quality}/plain"
        url = full_storage_url.replace(server_address, f"{server_address}/{resize_string}")
        return url
    except Exception as e:
        logger.debug(f"Failed to resize image with url: {full_storage_url}: {repr(e)}")
        return full_storage_url


def get_preview_link(title="preview"):
    return (
        f'<a href="javascript:;">{title}<i class="zmdi zmdi-cast" style="margin-left: 5px"></i></a>'
    )


def get_datetime(value: str) -> datetime:
    if value is None:
        return None
    return datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%fZ")


def get_readable_datetime(value: str) -> str:
    dt = get_datetime(value)
    if dt is None:
        return None
    return dt.strftime("%Y-%m-%d %H:%M:%S")


def get_unix_timestamp() -> int:
    """Return the current Unix timestamp.

    :return: Current Unix timestamp.
    :rtype: int
    """
    return int(time.time())


def get_certificates_list(path: str = DEFAULT_CA_BUNDLE_PATH) -> List[str]:
    with open(path, "r", encoding="ascii") as f:
        content = f.read().strip()
        certs = []

        begin_cert = "-----BEGIN CERTIFICATE-----"
        end_cert = "-----END CERTIFICATE-----"

        while begin_cert in content:
            start_index = content.index(begin_cert)
            end_index = content.index(end_cert, start_index) + len(end_cert)
            cert = content[start_index:end_index]
            certs.append(cert)
            content = content[end_index:]
        return certs


def setup_certificates():
    """
    This function is used to add extra certificates to the default CA bundle on Supervisely import.
    """
    path_to_certificate: str = os.environ.get("SLY_EXTRA_CA_CERTS", "").strip()
    if path_to_certificate == "":
        return

    if os.path.exists(path_to_certificate):
        if os.path.isfile(path_to_certificate):
            with open(path_to_certificate, "r", encoding="ascii") as f:
                extra_ca_contents = f.read().strip()
                if extra_ca_contents == "":
                    raise RuntimeError(f"File with certificates is empty: {path_to_certificate}")

            certificates = get_certificates_list(DEFAULT_CA_BUNDLE_PATH)
            requests_ca_bundle = os.environ.get("REQUESTS_CA_BUNDLE", "").strip()
            if requests_ca_bundle != "" and os.path.exists(requests_ca_bundle):
                if os.path.isfile(requests_ca_bundle):
                    certificates = get_certificates_list(requests_ca_bundle)
                else:
                    raise RuntimeError(f"Path to bundle is not a file: {requests_ca_bundle}")

            certificates.insert(0, extra_ca_contents)
            new_bundle_path = os.path.join(gettempdir(), "sly_extra_ca_certs.crt")
            with open(new_bundle_path, "w", encoding="ascii") as f:
                f.write("\n".join(certificates))

            old_request_ca_bundle_path = requests_ca_bundle
            os.environ["REQUESTS_CA_BUNDLE"] = new_bundle_path
            if (
                os.environ.get("SSL_CERT_FILE", "").strip() == ""
                or os.environ.get("SSL_CERT_FILE", "").strip() == old_request_ca_bundle_path
            ):
                os.environ["SSL_CERT_FILE"] = new_bundle_path
            logger.info(f"Certificates were added to the bundle: {path_to_certificate}")
        else:
            raise RuntimeError(f"Path to certificate is not a file: {path_to_certificate}")
    else:
        raise RuntimeError(f"Path to certificate does not exist: {path_to_certificate}")


def add_callback(func, callback):
    @wraps(func)
    def wrapper(*args, **kwargs):
        res = func(*args, **kwargs)
        callback()
        return res

    return wrapper


def compare_dicts(
    template: Dict[Any, Any], data: Dict[Any, Any], strict: bool = True
) -> Tuple[List[str], List[str]]:
    """Compare two dictionaries recursively (by keys only) and return lists of missing and extra fields.
    If strict is True, the keys of the template and data dictionaries must match exactly.
    Otherwise, the data dictionary may contain additional keys that are not in the template dictionary.

    :param template: The template dictionary.
    :type template: Dict[Any, Any]
    :param data: The data dictionary.
    :type data: Dict[Any, Any]
    :param strict: If True, the keys of the template and data dictionaries must match exactly.
    :type strict: bool, optional
    :return: A tuple containing a list of missing fields and a list of extra fields.
    :rtype: Tuple[List[str], List[str]]
    """
    missing_fields = []
    extra_fields = []

    if not isinstance(template, dict) or not isinstance(data, dict):
        return missing_fields, extra_fields

    if strict:
        template_keys = set(template.keys())
        data_keys = set(data.keys())

        missing_fields = list(template_keys - data_keys)
        extra_fields = list(data_keys - template_keys)

        for key in template_keys & data_keys:
            sub_missing, sub_extra = compare_dicts(template[key], data[key], strict)
            missing_fields.extend([f"{key}.{m}" for m in sub_missing])
            extra_fields.extend([f"{key}.{e}" for e in sub_extra])
    else:
        for key in template:
            if key not in data:
                missing_fields.append(key)
            else:
                sub_missing, sub_extra = compare_dicts(template[key], data[key], strict)
                missing_fields.extend([f"{key}.{m}" for m in sub_missing])
                extra_fields.extend([f"{key}.{e}" for e in sub_extra])

    return missing_fields, extra_fields


def get_or_create_event_loop() -> asyncio.AbstractEventLoop:
    """
    Get the current event loop or create a new one if it doesn't exist.
    Works for different Python versions and contexts.

    :return: Event loop
    :rtype: asyncio.AbstractEventLoop
    """
    try:
        # Preferred method for asynchronous context (Python 3.7+)
        return asyncio.get_running_loop()
    except RuntimeError:
        # If the loop is not running, get the current one or create a new one (Python 3.8 and 3.9)
        try:
            return asyncio.get_event_loop()
        except RuntimeError:
            # For Python 3.10+ or if the call occurs outside of an active loop context
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            return loop


def get_filename_from_headers(url):
    try:
        response = requests.head(url, allow_redirects=True)
        if response.status_code >= 400 or "Content-Disposition" not in response.headers:
            response = requests.get(url, stream=True)
        content_disposition = response.headers.get("Content-Disposition")
        if content_disposition:
            filename = re.findall('filename="?([^"]+)"?', content_disposition)
            if filename:
                return filename[0]
        filename = url.split("/")[-1] or "downloaded_file"
        return filename
    except Exception as e:
        print(f"Error retrieving file name from headers: {e}")
        return None
